"""merc_common URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/3.1/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  path('', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  path('', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.urls import include, path
    2. Add a URL to urlpatterns:  path('blog/', include('blog.urls'))
"""
from crams.views.auth import rapid_connect_auth_view
from crams.views.funding_body import PrevUserFundingBodyRoleList
from crams.views.lookup_api import EResearchBodySystemViewSet
from crams.views.lookup_api import fb_scheme_list
from crams.views.lookup_api import for_codes
from crams.views.lookup_api import SupportEmailContactViewSet

from django.contrib.auth import views as auth_views
from django.urls import include
from django.urls import path, re_path
from django.views.generic import TemplateView
from rest_framework import routers
from rest_framework.schemas import get_schema_view

router = routers.SimpleRouter()
router.register(r'e_research_body', EResearchBodySystemViewSet, basename='e_r_b')
router.register(r'user_funding_body', PrevUserFundingBodyRoleList, basename='funding_body')
router.register(r'support_email_list', SupportEmailContactViewSet, basename='support-emails')

urlpatterns = [
    # Use the `get_schema_view()` helper to add a `SchemaView` to project URLs.
    #   * `title` and `description` parameters are passed to `SchemaGenerator`.
    #   * Provide view name for use with `reverse()`.
    path('openapi', get_schema_view(
        title="Merc Common API",
        description="API for all things in Merc Common",
        version="1.0.0"
    ), name='openapi-schema'),
    # Route TemplateView to serve Swagger UI template.
    #   * Provide `extra_context` with view name of `SchemaView`.
    path('swagger-ui/', TemplateView.as_view(
        template_name='swagger-ui.html',
        extra_context={'schema_url': 'openapi-schema'}
    ), name='swagger-ui'),

    path('', include(router.urls)),
    re_path(r'^accounts/login/$', auth_views.LoginView.as_view()),
    re_path(r'^rapid_connect_token_auth', rapid_connect_auth_view),
    re_path(r'funding_scheme/(?P<fb_name>\w+)', fb_scheme_list, name='funding_scheme'),
    re_path(r'for_codes/', for_codes, name='for_codes'),
]



# Generated by Django 3.1.1 on 2020-09-25 02:27

from django.conf import settings
import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('crams', '0002_erb'),
    ]

    operations = [
        migrations.CreateModel(
            name='CramsSequence',
            fields=[
                ('code', models.CharField(max_length=255, primary_key=True, serialize=False)),
                ('last_sequence_num', models.IntegerField(default=0)),
            ],
        ),
        migrations.CreateModel(
            name='Duration',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('duration', models.IntegerField(default=1, validators=[django.core.validators.MinValueValidator(1)])),
                ('duration_label', models.CharField(max_length=50)),
            ],
        ),
        migrations.CreateModel(
            name='FundingBody',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=200)),
                ('email', models.EmailField(blank=True, max_length=254, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='Question',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('key', models.CharField(max_length=50, unique=True)),
                ('question_type', models.CharField(max_length=200)),
                ('question', models.CharField(max_length=200)),
            ],
        ),
        migrations.CreateModel(
            name='FORCode',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('code', models.CharField(max_length=50)),
                ('description', models.CharField(max_length=200)),
            ],
        ),
        migrations.CreateModel(
            name='SEOCode',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('code', models.CharField(max_length=200)),
                ('description', models.CharField(max_length=200)),
            ],
        ),
        migrations.CreateModel(
            name='FundingScheme',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('funding_scheme', models.CharField(max_length=200)),
                ('funding_body',
                 models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='funding_schemes',
                                   to='crams.fundingbody')),
            ],
        ),
    ]



# coding=utf-8
"""
    Python utils
"""


def has_invalid_prefix(str, invalid_prefix_list):
    return not validate_prefix(str, invalid_prefix_list, False)


def has_valid_prefix(str, valid_prefix_list):
    return validate_prefix(str, valid_prefix_list, True)


def validate_prefix(str, prefix_list, prefix_list_is_valid=True):
    for prefix in prefix_list:
        if str and str.startswith(prefix):
            return prefix_list_is_valid
    return not prefix_list_is_valid


def strip_lower(str):
    return str.strip().lower()


def reverse_dict(d):
    return {v: k for k, v in d.items()}


def update_return_dict(map, key, value):
    map.update({key: value})
    return map


# http://code.activestate.com/recipes/52308-the-simple-but-handy-collector-of-a-bunch-of-named/
class Bunch(list):

    """

    :param args:
    :param kw:
    """

    def __init__(self, *args, **kw):
        super(Bunch, self).__init__()
        self[:] = list(args)
        setattr(self, '__dict__', kw)


class BunchDict(dict):

    """

    :param kw:
    """

    def __init__(self, **kw):
        # noinspection PyTypeChecker
        dict.__init__(self, kw)
        self.__dict__ = self

    def __str__(self):
        state = ["%s=%r" % (attribute, value)
                 for (attribute, value)
                 in self.__dict__.items()]
        return '\n'.join(state)


# http://stackoverflow.com/questions/390250/elegant-ways-to-support-equivalence-equality-in-python-classes
class CommonEqualityMixin(object):
    """
        Common Equality Mixin
    """
    def __eq__(self, other):
        if isinstance(other, type(self)):
            return self.__dict__ == other.__dict__
        return False

    def __ne__(self, other):
        return not self.__eq__(other)


class SortableObject(CommonEqualityMixin):
    """

    :param obj:
    :param sort_lambda:
    """

    def __init__(self, obj, sort_lambda):
        self.obj = obj
        self.sortLambda = sort_lambda

    def is_same_internal_object(self, other):
        """

        :param other:
        :return:
        """
        if (isinstance(other, type(self))) or\
                (isinstance(self.obj, type(other))):
            return other.obj == self.obj
        return False



from crams import models
from crams.models import FORCode
from crams.models import FundingScheme
from crams.permissions import IsCramsAuthenticated
from crams.serializers import lookup_serializers
from crams.utils.model_lookup_utils import LookupDataModel
from rest_framework import mixins
from rest_framework import viewsets, decorators
from rest_framework.decorators import api_view
from rest_framework.views import Response


@api_view(http_method_names=['GET'])
def fb_scheme_list(request, fb_name):
    """
    <B>get Funding Scheme for given Funding Body, all if none</B>
    <BR><BR>param fb_name: (optional, case-insensitive) name of the funding body to filter results
    <BR><BR>return: a list of funding schemes, filtered by funding body name if given as parameter to API

    """
    qs = FundingScheme.objects.all()
    if fb_name:
        qs = qs.filter(funding_body__name__iexact=fb_name.lower())

    fs_list = []
    for fs in qs.order_by('id'):
        fs_list.append(lookup_serializers.FundingSchemeSerializer(fs).data)
    return Response(fs_list)


@api_view(http_method_names=['GET'])
def for_codes(request):
    """
    <B>get a list of FOR codes</B>
    <BR><BR>return: a list of FOR code json (Format: {'id':<db_id>, 'desc': '<for code> <description>'})
    """
    for_code_objs = FORCode.objects.all().order_by('code')

    for_codes_list = []
    for for_code in for_code_objs:
        code = for_code.code
        desc = for_code.description
        for_code_dict = {'id': for_code.id, 'desc': '{} {}'.format(code, desc)}
        for_codes_list.append(for_code_dict)
    return Response(for_codes_list)


class EResearchBodySystemViewSet(viewsets.ViewSet):
    serializer_class = None
    # permission_classes = [IsCramsAuthenticated]
    queryset = models.EResearchBody.objects.all()

    @decorators.action(detail=False, url_path='list')
    def fetch_e_research_body(self, request):
        """
        <B>Fetch a list of available eResearch Body</B>
        <BR><BR>:return: a list of eResearch Body
        """
        ret_list = []
        for body in models.EResearchBody.objects.all():
            ret_list.append(body.name)

        return Response(ret_list)

    @decorators.action(detail=False, url_path='e_research_system/(?P<e_research_body>[-\w]+)')
    def fetch_e_research_systems(self, request, e_research_body):
        """
        <B>Given an e_research_body name, this api returns a list of eResearch Systems belonging to it.</B>
        <BR><BR>param: e_research_body: string value of e_research_body name, case insensitive.
        <BR><BR>return: a list of eResearch Systems belonging to the input eResearch body

        """
        ret_list = []
        lookup = LookupDataModel(models.EResearchBody)
        serializer_class = lookup_serializers.EResearchSystemSerializer
        try:
            search_key = {'name__iexact': e_research_body}
            body_obj = lookup.get_lookup_data(search_key)
            qs = body_obj.systems.all()
            ret_list = serializer_class(qs, many=True).data
        except Exception as e:
            pass

        return Response(ret_list)


class SupportEmailContactViewSet(viewsets.GenericViewSet, mixins.ListModelMixin):
    permission_classes = [IsCramsAuthenticated]
    serializer_class = lookup_serializers.SupportEmailContactSerializer
    queryset = models.SupportEmailContact.objects.all()

    def get_queryset(self):
        queryset = self.queryset
        if 'erb' in self.request.query_params:
            erb = self.request.query_params['erb']
            queryset = self.queryset.filter(e_research_body__name=erb)

        if 'erbs' in self.request.query_params:
            erbs = self.request.query_params['erbs']
            queryset = self.queryset.filter(e_research_system__name=erbs)

        return queryset



# coding=utf-8
"""Lookup Serializers."""
from django.db.models import Q
from django.contrib.auth import get_user_model
from rest_framework import serializers, exceptions

# from crams.models import AllocationHome
# from crams.models import ComputeProduct
# from crams.models import ContactRole
# from crams.models import FORCode
from crams.models import FundingBody
from crams.models import FundingScheme
# from crams.models import GrantType
from crams.models import EResearchBodyIDKey
from crams.models import Provider
from crams.models import Question
# from crams.models import RequestStatus
# from crams.models import StorageProduct
# from crams.models import StorageType
from crams.models import SupportEmailContact
from crams.models import Zone
from crams.models import EResearchBodySystem, EResearchBody

from crams.serializers import model_serializers

User = get_user_model()


class QuestionSerializer(model_serializers.ModelLookupSerializer):
    """class QuestionSerializer."""

    id = serializers.IntegerField(required=False)

    key = serializers.CharField()

    question_type = serializers.CharField(required=False)

    question = serializers.CharField(required=False)

    class Meta(object):
        """metaclass."""

        model = Question
        field = ('id', 'key', 'question_type', 'question')
        pk_fields = ['key']
        read_only_fields = ('id', 'question_type', 'question')

    def validate(self, attrs):
        attrs = super().validate(attrs)
        return self.instance or attrs


class ProviderSerializer(serializers.ModelSerializer):
    """class ProviderSerializer."""

    class Meta(object):
        """metaclass."""

        model = Provider
        fields = ('id', 'name', 'active')
        # read_only_fields = ('id', 'name')


# class RequestStatusSerializer(serializers.ModelSerializer):
#     """class RequestStatusSerializer."""
#
#     class Meta(object):
#         """metaclass."""
#
#         model = RequestStatus
#         fields = ('id', 'code', 'status')
#

class FundingBodySchemeSerializer(serializers.ModelSerializer):
    """class FundingBodySchemeSerializer"""
    funding_body = serializers.SlugRelatedField(
        slug_field='name', queryset=FundingBody.objects.all())

    class Meta(object):
        """metaclass."""

        model = FundingScheme
        fields = ('funding_scheme', 'funding_body')


class FundingSchemeSerializer(serializers.ModelSerializer):
    """class FundingSchemeSerializer."""

    class Meta(object):
        """metaclass."""

        model = FundingScheme
        fields = ('id', 'funding_scheme')


class FundingBodySerializer(serializers.ModelSerializer):
    """class FundingBodySerializer."""

    class Meta(object):
        """metaclass."""

        model = FundingBody
        fields = ('id', 'name')
        read_only_fields = ('id', 'name')


# class ComputeProductSerializer(serializers.ModelSerializer):
#     """class ComputeProductSerializer."""
#
#     funding_body = FundingBodySerializer(many=False, read_only=True)
#     provider = ProviderSerializer(many=False, read_only=True)
#
#     class Meta(object):
#         """metaclass."""
#
#         model = ComputeProduct
#         fields = ('id', 'name', 'funding_body', 'provider')
#         read_only_fields = ('funding_body', 'provider')
#
#
# class StorageTypeSerializer(serializers.ModelSerializer):
#     """class StorageTypeSerializer."""
#
#     class Meta(object):
#         """metaclass."""
#
#         model = StorageType
#         fields = ('id', 'storage_type')


class ZoneSerializer(serializers.ModelSerializer):
    """class ZoneSerializer."""

    class Meta(object):
        """metaclass."""

        model = Zone
        fields = ('id', 'name')


# class ParentStorageProductSerializer(serializers.ModelSerializer):
#     class Meta(object):
#         model = StorageProduct
#         fields = ('id', 'name')
#
#
# class AllocationComputeProductValidate(model_serializers.ModelLookupSerializer):
#     id = serializers.IntegerField()
#     name = serializers.CharField(required=False)

#     class Meta(object):
#         """metaclass."""
#         model = ComputeProduct
#         fields = ('id', 'name')

#     def validate(self, attrs):
#         super().validate(attrs)
#         self.verify_field_value('name', attrs, ignore_case=True)
#         return attrs
#
#
# class StorageProductZoneOnlySerializer(model_serializers.
#                                        ModelLookupSerializer):
#     """class StorageProductZoneOnlySerializer."""
#     id = serializers.IntegerField()
#     storage_type = StorageTypeSerializer(many=False, read_only=True)
#     zone = ZoneSerializer(many=False, read_only=True, required=False)
#     name = serializers.CharField(required=False)
#     parent_storage_product = ParentStorageProductSerializer(
#         many=False, read_only=True)
#
#     class Meta(object):
#         """metaclass."""
#         model = StorageProduct
#         fields = ('id', 'name', 'storage_type',
#                   'zone', 'parent_storage_product')
#         read_only_fields = ('storage_type', 'zone', 'parent_storage_product')
#
#     def validate(self, attrs):
#         super().validate(attrs)
#         self.verify_field_value('name', attrs, ignore_case=True)
#         return attrs
#
#
# class StorageProductSerializer(serializers.ModelSerializer):
#     """class StorageProductSerializer."""
#
#     storage_type = StorageTypeSerializer(many=False, read_only=True)
#
#     funding_body = FundingBodySerializer(many=False, read_only=True)
#
#     provider = ProviderSerializer(many=False, read_only=True)
#
#     zone = ZoneSerializer(many=False, read_only=True)
#
#     name = serializers.CharField(required=False)
#
#     parent_storage_product = ParentStorageProductSerializer(many=False,
#                                                             read_only=True)
#
#     class Meta(object):
#         """metaClass."""
#
#         model = StorageProduct
#         fields = ('id', 'name', 'storage_type',
#                   'funding_body', 'provider', 'zone', 'parent_storage_product')
#
#
# class GrantTypeSerializer(serializers.ModelSerializer):
#     """class GrantTypeSerializer."""
#
#     class Meta(object):
#         """metaclass."""
#
#         model = GrantType
#         fields = ('id', 'description')
#

class EResearchBodyIDKeySerializer(serializers.ModelSerializer):
    """class EResearchBodyIDKeySerializer."""

    class Meta(object):
        """metaclass."""

        model = EResearchBodyIDKey
        fields = ('id', 'key')


#
#
# class ContactRoleSerializer(serializers.ModelSerializer):
#     """class ContactRoleSerializer."""
#
#     class Meta(object):
#         """metaclass."""
#
#         model = ContactRole
#         field = ('id', 'name')


class SupportEmailContactSerializer(serializers.ModelSerializer):
    """Class SupportEmailContactSerializer."""
    e_research_body = serializers.SerializerMethodField()
    e_research_system = serializers.SerializerMethodField()

    class Meta(object):
        """metaclass."""

        model = SupportEmailContact
        fields = ('id', 'description', 'email',
                  'e_research_body', 'e_research_system')

    @classmethod
    def get_e_research_body(cls, obj):
        return obj.e_research_body.name

    @classmethod
    def get_e_research_system(cls, obj):
        if obj.e_research_system:
            return obj.e_research_system.name
        else:
            return None


#
#
# class FORCodeSerializer(serializers.ModelSerializer):
#     """class FORCodeSerializer."""
#
#     class Meta(object):
#         """metaclass."""
#
#         model = FORCode
#         fields = ('id', 'code', 'description')
#
#
# class AllocationHomeSerializer(serializers.ModelSerializer):
#     """class AllocationHomeSerializer."""
#
#     class Meta(object):
#         """metaclass."""
#
#         model = AllocationHome
#         fields = ('id', 'code', 'description')


class UserSerializer(serializers.ModelSerializer):
    """class UserSerializer."""

    class Meta(object):
        """metaclass."""

        model = User
        fields = ('first_name', 'last_name', 'email')  # , 'auth_token')


class EResearchBodySerializer(serializers.ModelSerializer):
    """class EResearchBodySerializer."""

    class Meta(object):
        """metaclass."""

        model = EResearchBody
        fields = ('name')
        read_only_fields = ('name')


class EResearchSystemSerializer(model_serializers.ReadOnlyModelSerializer):
    """class EResearchSystemSerializer."""
    e_research_body = serializers.SlugRelatedField(
        slug_field='name', queryset=EResearchBody.objects.all())

    class Meta(object):
        """metaclass."""

        model = EResearchBodySystem
        validators = []
        fields = ('id', 'name', 'e_research_body')

    def fetch_object(self):
        try:
            return EResearchBodySystem.objects.get(
                name=self.name, e_research_body=self.e_research_body)
        except Exception as e:
            pass
        return None
#
#
# class ERBSystemComputeProductSerializer(model_serializers.
#                                         ReadOnlyModelSerializer):
#     """class ComputeProductSerializer."""
#
#     e_research_system = EResearchSystemSerializer()
#
#     funding_body = \
#         serializers.SlugRelatedField(slug_field='name', read_only=True)
#
#     class Meta(object):
#         """metaclass."""
#
#         model = ComputeProduct
#         validators = []
#         fields = ('id', 'name', 'e_research_system', 'funding_body')
#
#
# class EResearchBodyParentStorageProductSerializer(model_serializers.
#                                                   ReadOnlyModelSerializer):
#     class Meta(object):
#         model = StorageProduct
#         fields = ('id', 'name')
#
#
# class ERBSystemStorageProductSerializer(model_serializers.
#                                         ReadOnlyModelSerializer):
#     e_research_system = EResearchSystemSerializer()
#
#     storage_type = serializers.SlugRelatedField(
#         slug_field='storage_type', read_only=True)
#
#     zone = serializers.SlugRelatedField(slug_field='name', read_only=True)
#
#     parent_storage_product = EResearchBodyParentStorageProductSerializer(
#         many=False, read_only=True)
#
#     class Meta(object):
#         model = StorageProduct
#         validators = []
#         fields = ('id', 'name', 'e_research_system', 'zone',
#                   'storage_type', 'funding_body', 'parent_storage_product')



# coding=utf-8
"""
Common ERB related type Serializers
"""

from django.db import transaction

from crams import models
from crams.utils.role import AbstractCramsRoleUtils
from rest_framework import serializers, exceptions
from crams.serializers import model_serializers
from crams.utils import rest_utils
from abc import abstractmethod


class EResearchBodyIDKeySerializer(model_serializers.ModelLookupSerializer):
    """class EResearchBodyIDKeySerializer."""
    id = serializers.IntegerField(required=False)

    key = serializers.CharField(required=False)

    e_research_body = serializers.SlugRelatedField(
        slug_field='name', required=False,
        queryset=models.EResearchBody.objects.all())

    type = serializers.SerializerMethodField()

    class Meta(object):
        model = models.EResearchBodyIDKey
        field = ('id', 'key', 'e_research_body', 'type')
        pk_fields = ['key', 'e_research_body']
        flags = [model_serializers.ModelLookupSerializer.IGNORE_NOT_EXIST]

    @classmethod
    def build_label_list_for_erb(cls, erb_obj):
        return cls.build_obj_list_for_type(
            erb_obj, obj_type=models.EResearchBodyIDKey.LABEL)

    @classmethod
    def build_metadata_list_for_erb(cls, erb_obj):
        return cls.build_obj_list_for_type(
            erb_obj, obj_type=models.EResearchBodyIDKey.METADATA)

    @classmethod
    def build_idkey_list_for_erb(cls, erb_obj):
        return cls.build_obj_list_for_type(
            erb_obj, obj_type=models.EResearchBodyIDKey.LABEL)

    @classmethod
    def build_obj_list_for_type(cls, erb_obj, obj_type):
        ret_list = list()
        if isinstance(erb_obj, models.EResearchBody):
            qs = models.EResearchBodyIDKey.objects.filter(
                e_research_body=erb_obj, type=obj_type)
            for label in qs:
                ret_list.append(label.key)
        return ret_list

    @classmethod
    def get_type(cls, obj):
        return obj.get_type_display()

    def validate(self, attrs):
        attrs = super().validate(attrs)
        self.verify_field_value('key', attrs, ignore_case=True)
        self.verify_field_value('e_research_body', attrs)
        return self.instance or attrs

    def create_new_metadata(self):
        self.validated_data['type'] = models.EResearchBodyIDKey.METADATA
        return self.create_new()

    # def create_new_label(self):
    #     self.validated_daMETADATAta['type'] = models.EResearchBodyIDKey.LABEL
    #     return self.create_new()

    def create_new_idkey(self):
        self.validated_data['type'] = models.EResearchBodyIDKey.ID_KEY
        return self.create_new()

    # def create_new_for_type(self, obj_type):
    #     if obj_type == models.EResearchBodyIDKey.ID_KEY:
    #         return self.create_new_idkey()
    #     if obj_type == models.EResearchBodyIDKey.METADATA:
    #         return self.create_new_metadata()
    #     return self.create_new_label()

    @transaction.atomic
    def create_new(self):
        if not self.validated_data:
            self.is_valid(raise_exception=True)

        if self.instance:
            erb_obj = self.instance.e_research_body
        else:
            erb_obj = self.validated_data.get('e_research_body')

        # erb label can only be created/updated by admin:
        user, context = rest_utils.get_current_user_from_context(self)
        if not AbstractCramsRoleUtils.is_user_erb_admin(
                user_obj=user, erb_obj=erb_obj, include_systems=True):
            msg = 'User not authorised to add label for erb: {}'
            raise exceptions.ValidationError(msg.format(erb_obj.name))

        if self.instance:
            return self.instance

        description = self.initial_data.get('description')
        if description:
            self.validated_data['description'] = description

        return self.Meta.model.objects.create(**self.validated_data)


class BaseIdentifierSerializer(model_serializers.ReadOnlyModelSerializer):
    system = EResearchBodyIDKeySerializer()

    class Meta(object):
        """meta Class."""

        model = None
        fields = ('identifier', 'system')

    # noinspection PyUnusedLocal
    @classmethod
    def _init_empty(cls, str_val, num):
        ret_dict = dict()
        ret_dict['identifier'] = 'System Identifier'
        ret_dict['system'] = {'e_research_body': 'ERB Name',
                              'key': 'RANDOM_KEY'
                              }
        return ret_dict

    @classmethod
    def _init_from_project_id_data(cls, project_id_data):
        ret_dict = dict()
        ret_dict['identifier'] = project_id_data.get('identifier', None)
        ret_dict['system'] = project_id_data.get('system', None)
        return ret_dict

    @classmethod
    def fetch_identifier_unique_for_system(cls, search_data, model):
        try:
            return model.objects.get(**search_data)
        except model.DoesNotExist:
            pass
        except model.MultipleObjectsReturned:
            msg = 'Unable for find unique {} identifer for {}'
            raise exceptions.ValidationError(msg.format(
                model.__class__, search_data))

    @abstractmethod
    def get_user_erb_roles(self):
        #
        # role_fn = user_utils.fetch_erb_userroles_with_provision_privileges
        # user, context = user_utils.get_current_user_from_context(self)
        return []

    def validate_system(self, system_obj):
        if not isinstance(system_obj, models.EResearchBodyIDKey):
            msg = 'Unable to determine system object for {}'
            raise exceptions.ValidationError(msg.format(system_obj))

        erb_required = system_obj.e_research_body

        valid_user = False
        for user_erb_role in self.context.get(
                'user_erb_roles', self.get_user_erb_roles()):
            if user_erb_role.role_erb == erb_required:
                valid_user = True
                break

        is_clone = self.context and 'CLONE' in self.context
        if not valid_user and not is_clone:
            msg = 'User does hold permission to update identifier for {}'
            raise exceptions.ValidationError(msg.format(erb_required))

        return system_obj



# coding=utf-8
"""

"""

from rest_framework import exceptions

from crams import models


class ProvisionDetailUtils:
    def __init__(self):
        pass

    @classmethod
    def get_provisioned(cls, obj):
        p_details = obj.provision_details
        provisioned_list = [models.ProvisionDetails.PROVISIONED]
        return p_details and p_details.status in provisioned_list

    @classmethod
    def validate_provisioned(cls, provisioned_status, instance, msg_param,
                             is_clone=False, is_admin=False):

        if provisioned_status not in [True, False]:
            msg = 'provisioned value must be boolean, not {}'
            raise exceptions.ValidationError(
                msg.format(provisioned_status.__class__.__name__))

        if is_clone or is_admin:
            return

        msg = 'not sent for provisioning, hence cannot be update ' \
              'Provision status'
        if instance:
            pd = instance.provision_details
            if pd:
                if pd.status in models.ProvisionDetails.SET_OF_SENT:
                    msg = 'cannot be updated unless provision status ' \
                          'is set to true'
                    if provisioned_status:
                        return
                if pd.status == models.ProvisionDetails.PROVISIONED:
                    msg = 'is currently provisioned, cannot be updated'
        raise exceptions.ValidationError(msg_param + ' ' + msg)

    @classmethod
    def build_new_provision_details(
            cls, instance, validated_data, current_user):
        message = validated_data.get('message', None)

        status = models.ProvisionDetails.UNKNOWN
        # create copy of previous provision details
        if instance.provision_details:
            old_pd = instance.provision_details
            if not message:
                message = old_pd.message
            status = old_pd.status

        provisioned = validated_data.pop('provisioned')
        if provisioned:
            status = models.ProvisionDetails.PROVISIONED
            message = None

        return models.ProvisionDetails(
            status=status, message=message, created_by=current_user)

    @classmethod
    def get_sent_object(cls, message=None):
        status = models.ProvisionDetails.SENT
        return cls.build_object(status, message)

    @classmethod
    def validate_existing_details_obj(
            cls, existing_details_obj, expected_status_set):
        if not isinstance(existing_details_obj, models.ProvisionDetails):
            msg = 'Provision Details sent instance expected, got {}'.format(
                type(existing_details_obj))
            raise exceptions.ParseError(msg)

        if not existing_details_obj.id:
            msg = 'to validate, Provision Details must be an existing object'
            raise exceptions.ParseError(msg)

        status = existing_details_obj.status
        if status not in expected_status_set:
            msg = 'Provision Details status must be one of {} got {}'.format(
                expected_status_set, status)
            raise exceptions.ParseError(msg)

    @classmethod
    def get_post_provision_update_sent_object(
            cls, existing_details_obj, message=None):
        expected_set = set(models.ProvisionDetails.POST_PROVISION_UPDATE)
        cls.validate_existing_details_obj(existing_details_obj, expected_set)
        status = models.ProvisionDetails.POST_PROVISION_UPDATE_SENT
        return cls.build_object(status, message)

    @classmethod
    def get_post_provision_update_object(
            cls, existing_details_obj, message=None):
        cls.validate_existing_details_obj(
            existing_details_obj, set(models.ProvisionDetails.PROVISIONED))
        status = models.ProvisionDetails.POST_PROVISION_UPDATE
        return cls.build_object(status, message)

    @classmethod
    def get_provisioned_object(cls, provisioned_details_obj, message=None):
        cls.validate_existing_details_obj(
            provisioned_details_obj, models.ProvisionDetails.SET_OF_SENT)
        status = models.ProvisionDetails.PROVISIONED
        return cls.build_object(status, message)

    @classmethod
    def get_failed_object(cls, sent_details_obj, message):
        cls.validate_existing_details_obj(
            sent_details_obj, models.ProvisionDetails.SET_OF_SENT)
        status = models.ProvisionDetails.FAILED
        return cls.build_object(status, message)

    @classmethod
    def build_object(cls, status, message):
        # Do not save here, just leave it in memory
        return models.ProvisionDetails(status=status, message=message)

    @classmethod
    def update_or_get_new_provisiondetails_object(
            cls, provisionable_object, status_boolean):

        p_status = models.ProvisionDetails.SENT
        p_message = None
        if status_boolean:
            p_status = models.ProvisionDetails.PROVISIONED
            p_message = None

        existing_provision_detail = provisionable_object.provision_details
        if existing_provision_detail:
            if not status_boolean and existing_provision_detail.status \
                    == models.ProvisionDetails.PROVISIONED:
                msg = 'Cannot de-provision a provisioned identifier'
                raise exceptions.ValidationError(msg)
            existing_provision_detail.status = p_status
            existing_provision_detail.message = None
            # Do not save here, just leave it in memory
            return existing_provision_detail

        return cls.build_object(status=p_status, message=p_message)



from django.db import migrations, models

def load_crams_member_inital_data_from_sql():
    import os

    file_name = 'project_member_status.sql'
    sql_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), '../sqls'))
    sql_statements = open(os.path.join(sql_dir, file_name), 'r').read()

    return sql_statements

class Migration(migrations.Migration):

    dependencies = [
        ('crams', '0009_projectmemberstatus'),
    ]

    operations = [
        migrations.RunSQL(load_crams_member_inital_data_from_sql()),
    ]





# coding=utf-8
"""model related serializers"""

import ast

from rest_framework import serializers, exceptions as rest_exceptions
from django.core.exceptions import ObjectDoesNotExist
from django.core.exceptions import ValidationError as django_ValidationError
from django.db import models, transaction

from crams import models as crams_models
# from crams import permissions
from crams.utils import debug, rest_utils
# from crams.constants import code


# class HasPermissionModelSerializer(serializers.Serializer):
#     def validate(self, attrs):
#         suffix_msg = ' required to determine save access'
#
#         error_message = '"context" object not found,' + suffix_msg
#         if not self.context or 'request' not in self.context:
#             raise Exception(error_message)
#
#         # if code.USER_HAS_SAVE_PROJECT_ACCESS not in self.context:
#         #     project = attrs.get('project')
#         #     if not project:
#         #         request = attrs.get('request')
#         #         if not request:
#         #             raise Exception(
#         #                 'Unable to identify associated Project ' + suffix_msg)
#         #         project = request.project
#         #
#         #         self.context[code.USER_HAS_SAVE_PROJECT_ACCESS] = \
#         #             permissions.has_save_access_to_project(
#         #                 project, self.context.request.user)
#         #
#         # no_access_msg = 'User does not have save access to associated Project'
#         # if not self.context[code.USER_HAS_SAVE_PROJECT_ACCESS]:
#         #     raise Exception(no_access_msg)
#
#         return attrs
#
#
# class CreateOnlySerializer(serializers.Serializer):
#     id = serializers.IntegerField(read_only=True)
#
#     def update(self, instance, validated_data):
#         raise rest_exceptions.ParseError('Update not allowed ')
#
#     def create(self, validated_data):
#         raise rest_exceptions.ParseError('create Not implemented yet ')


class CreateOnlyModelSerializer(serializers.ModelSerializer):
    id = serializers.IntegerField(read_only=True)

    def update(self, instance, validated_data):
        """update.

        :param instance:
        :param validated_data:
        """
        raise rest_exceptions.ParseError('Update not allowed ')

    def __delete__(self, instance):
        raise rest_exceptions.APIException('Delete not Allowed')

    def get_current_user(self):
        user, context = rest_utils.get_current_user_from_context(self)
        return user

    @classmethod
    def add_related_entities(
            cls, validated_data, related_tables_map, parent_set_callback_fn):
        for key, model in related_tables_map.items():
            data = validated_data.get(key)
            if data:
                for instance_data in data:
                    instance = model(**instance_data)
                    parent_set_callback_fn(instance)
                    instance.save()


class ReadOnlyModelSerializer(CreateOnlyModelSerializer):
    def create(self, validated_data):
        """
        create
        :param validated_data:
        :return:
        """
        raise rest_exceptions.ParseError('Create not allowed ')


class PrimaryKeyLookupField(serializers.PrimaryKeyRelatedField):
    """class PrimaryKeyLookupField.
        - cannot be used for optional serializer fields
    """

    def __init__(self, *args, **kwargs):
        """init.

        :param *args:
        :param **kwargs:
        """
        # Don't pass the 'fields' arg up to the superclass
        # Grant._meta.get_all_field_names()
        self.fields = kwargs.pop('fields', ['id'])
        self.pkFields = kwargs.pop('keyFields', ['id'])
        self.model = None

        # make sure this is a model class, not a string
        model = kwargs.pop('model', None)
        if model:
            msg = '{} is not an instance of django Model'.format(model)
            if not isinstance(model(), models.Model):
                raise rest_exceptions.ParseError(msg)
            self.model = model

        # Instantiate the superclass normally
        super(PrimaryKeyLookupField, self).__init__(*args, **kwargs)

    def to_internal_value(self, data):
        """to internal value.

        :param data:
        :return data:
        """
        if len(self.pkFields) > 0:
            data_dict = data
            if isinstance(data_dict, str):
                data_dict = ast.literal_eval(data)
            if not isinstance(data_dict, dict):
                msg = '{} is not a dict object'
                raise rest_exceptions.ParseError(msg, data)
            ret_dict = {}
            for k in self.pkFields:
                try:
                    ret_dict[k] = data_dict[k]
                except Exception:
                    raise rest_exceptions.ParseError(
                        '{}: Field {} required'.format(self.label, k))
            return self.get_queryset().get(**ret_dict)

        return super().to_internal_value(data)

    def to_representation(self, value):
        """to representation.

        :param value:
        :return ret_dict:
        """
        if self.pk_field is not None:
            return self.pk_field.to_representation(value.pk)

        try:
            if self.model:
                instance = self.model.objects.get(pk=value.pk)
            else:
                if isinstance(value, dict):
                    instance = self.get_queryset().get(**value)
                else:
                    instance = self.get_queryset().get(pk=value.pk)

            ret_dict = {}
            if self.fields:
                for k in self.fields:
                    if isinstance(k, dict):
                        for name, path in k.items():
                            node = instance
                            for attr in path.split('.'):
                                node = getattr(node, attr)
                            ret_dict[name] = node
                    else:
                        ret_dict[k] = getattr(instance, k)
            return ret_dict

        except ObjectDoesNotExist:
            self.fail('does_not_exist', pk_value=value.pk)
        except (TypeError, ValueError):
            self.fail('incorrect_type', data_type=type(value.pk).__name__)


class ModelLookupSerializer(serializers.Serializer, debug.Debug):
    IGNORE_NOT_EXIST = 'ignore_not_exist'

    def get_model_name(self):
        model_name = None
        if hasattr(self, 'Meta'):
            model_name = self.Meta.model.__name__
        return model_name

    def verify_field_value(self, field_name, attrs, ignore_case=False):
        if self.instance and field_name in attrs:
            field_value = attrs.get(field_name)
            msg = 'value for {} does not match {} with id {}'
            attr_value = getattr(self.instance, field_name)
            if ignore_case:
                field_value = field_value.lower()
                attr_value = attr_value.lower()
            if not field_value == attr_value:
                model_name = self.get_model_name()
                raise rest_exceptions.ValidationError(
                    msg.format(field_name, model_name, self.instance.id))

    def validate(self, attrs):
        if hasattr(self, 'Meta'):
            pk_dict = dict()
            pk_found = False
            if 'id' in attrs:
                pk = attrs.pop('id')
                if pk:
                    pk_dict['id'] = pk
                    pk_found = True
            else:
                if hasattr(self.Meta, 'pk_fields'):
                    pk_found = True
                    for pk_field in self.Meta.pk_fields:
                        if pk_field not in attrs:
                            pk_found = False
                            break
                        pk_dict[pk_field] = attrs.get(pk_field)

            if pk_found:
                if isinstance(self.Meta.model, crams_models.ArchivableModel):
                    if 'archive_ts__isnull' not in pk_dict:
                        pk_dict['archive_ts__isnull'] = True
                try:
                    self.instance = self.Meta.model.objects.get(**pk_dict)
                    invalid_fields = list()
                    for k, v in attrs.items():
                        if not hasattr(self.instance, k):
                            invalid_fields.append(k)

                    if invalid_fields:
                        raise rest_exceptions.ValidationError(
                            'Fields {} invalid for model {}'.format(
                                invalid_fields, self.Meta.model.__name__))

                except self.Meta.model.DoesNotExist:
                    if 'id' not in pk_dict and hasattr(self.Meta, 'flags'):
                        if self.IGNORE_NOT_EXIST in self.Meta.flags:
                            pass
                    else:
                        msg = '{} does not exist for {}'
                        raise rest_exceptions.ValidationError(
                            msg.format(self.Meta.model.__name__, pk_dict))
        return attrs

    def create(self, validated_data):
        msg = 'Create not allowed for {}'.format(self.get_model_name())
        raise rest_exceptions.ValidationError(msg)

    def update(self, instance, validated_data):
        msg = 'Update not allowed for {}'.format(self.get_model_name())
        raise rest_exceptions.ValidationError(msg)


class CreateModelLookupSerializer(ModelLookupSerializer):
    def update_related(self, instance, validated_data):
        error_dict = dict()
        update_nested = list()
        if hasattr(self.Meta, 'update_nested'):
            update_nested = self.Meta.update_nested

        for related_field in self.Meta.related_fk:
            if related_field not in validated_data:
                continue
            validated_data.pop(related_field)
            sz = self.fields.get(related_field)
            msg = 'Serializer not found for {}'.format(related_field)
            if not sz:
                raise rest_exceptions.ValidationError(msg)

            if related_field not in update_nested:
                validated_data[related_field] = sz.instance
                continue

            # update Nested relation data, if required
            related_data_initial = self.initial_data.get(related_field)
            new_sz = type(sz)(instance=sz.instance, data=related_data_initial)
            new_sz.is_valid(raise_exception=True)
            new_sz.save()
            validated_data[related_field] = new_sz.instance

        if error_dict:
            raise rest_exceptions.ValidationError(error_dict)

        return instance

    @classmethod
    def model_validate_save(cls, model_instance_obj):
        model_name = model_instance_obj._meta.model_name
        try:
            model_instance_obj.validate_unique()
            model_instance_obj.save()
        except django_ValidationError as e:
            msg = ','.join(e.messages)
            raise rest_exceptions.ValidationError({model_name: msg})
        except Exception as e:
            raise rest_exceptions.ValidationError({model_name: str(e)})

    def validate(self, data):
        data = super().validate(data)
        if self.instance:
            return data
        msg_base = 'field is required'
        err_dict = dict()
        for required in self.Meta.required_at_save:
            if required not in data:
                err_dict[required] = msg_base.format(required)
        if err_dict:
            raise rest_exceptions.ValidationError(err_dict)

        return data

    def save_common(self, instance, validated_data):
        # create related fields
        many_to_many_dict = dict()
        if hasattr(self.Meta, 'many_to_many'):
            base_msg = 'Many to Many field {} not found for {}'
            for many_field in self.Meta.many_to_many:
                msg = base_msg.format(many_field, self.Meta.model)
                if not hasattr(self.Meta.model, many_field):
                    raise rest_exceptions.ValidationError(msg)
                if many_field not in validated_data:
                    continue
                many_data_list = validated_data.pop(many_field)
                if many_data_list:
                    many_to_many_dict[many_field] = set(many_data_list)

        reverse_dict = dict()
        if hasattr(self.Meta, 'reverse_fk'):
            for (reverse_field, instance_field) in self.Meta.reverse_fk:
                if reverse_field not in validated_data:
                    continue
                field_validated_list = validated_data.pop(reverse_field)
                if not field_validated_list:
                    continue

                reverse_sz = self.fields.get(reverse_field)
                error_list = list()
                msg = 'Serialzer not found for {}'.format(reverse_field)
                if not reverse_sz:
                    error_list.append(msg)
                else:
                    msg = 'reverse serializer'
                    list_msg = msg + ' is not a list serializer'
                    if not reverse_sz.child:
                        error_list.append(list_msg)
                if error_list:
                    raise rest_exceptions.ValidationError({
                        reverse_field: error_list
                    })

                sz_class = type(reverse_sz.child)
                reverse_field_sz_list = list()
                for reverse_data in self.initial_data.get(reverse_field):
                    sz = sz_class(data=reverse_data)
                    sz.is_valid(raise_exception=True)
                    reverse_field_sz_list.append(sz)
                reverse_dict[reverse_field] = \
                    (reverse_field_sz_list, instance_field)

        if hasattr(self.Meta, 'related_fk'):
            self.update_related(instance, validated_data)

        if instance or self.instance:
            if not instance:
                instance = self.instance
            for k, v in validated_data.items():
                setattr(instance, k, v)
        else:
            instance = self.Meta.model(**validated_data)
        self.model_validate_save(instance)

        if reverse_dict:
            error_list = list()
            update_sz_list = list()
            for reverse_field, val in reverse_dict.items():
                (reverse_field_sz_list, instance_field) = val
                for sz in reverse_field_sz_list:
                    if sz.instance:
                        existing = getattr(sz.instance, instance_field)
                        if not existing == instance:
                            msg = 'Cannot change reverse parent for id'
                            error_list.append(msg.format(reverse_sz.instance))
                if error_list:
                    raise rest_exceptions.ValidationError({
                        reverse_field: error_list
                    })
                else:
                    sz.validated_data[instance_field] = instance
                    update_sz_list.append(sz)

            for sz in update_sz_list:
                if sz.instance:
                    changed = False
                    for k, v in sz.validated_data.items():
                        e_v = getattr(sz.instance, k)
                        if not v == e_v:
                            setattr(sz.instance, k, v)
                            changed = True
                    if changed:
                        sz.instance.save()
                    continue
                reverse_obj = sz.Meta.model(**sz.validated_data)
                self.model_validate_save(reverse_obj)

        if many_to_many_dict:
            error_dict = dict()
            for many_field, many_field_set in many_to_many_dict.items():
                try:
                    getattr(instance, many_field, many_field_set)
                except Exception as e:
                    error_dict[many_field] = str(e)
            if error_dict:
                raise rest_exceptions.ValidationError(error_dict)

        return instance

    @transaction.atomic()
    def create(self, validated_data):
        return self.save_common(None, validated_data)

    @transaction.atomic()
    def update(self, instance, validated_data):
        raise rest_exceptions.ValidationError('Update Not Implemented Yet')


class RelatedModelLookupSerializer(ModelLookupSerializer):
    pass


class UpdateModelLookupSerializer(CreateModelLookupSerializer):
    @transaction.atomic()
    def update(self, instance, validated_data):
        return self.save_common(instance, validated_data)



# coding=utf-8
"""
Mail Sender methods
"""
from collections import Iterable

from django.template import Context
from django.template.loader import get_template
from django.core.mail import EmailMessage

from django.conf import settings
import threading

__author__ = 'simonyu, rafi feroze, melvin luong'


class EmailThread(threading.Thread):
    def __init__(self, sender, subject, mail_content, recipient_list, template_name=None, cc_list=None, bcc_list=None,
                 reply_to=None, fail_silently=False):
        self.sender = sender
        self.subject = subject
        self.mail_content = mail_content
        self.recipient_list = recipient_list
        self.template_name = template_name
        self.cc_list = cc_list
        self.bcc_list = bcc_list
        self.reply_to = reply_to
        self.fail_silently = fail_silently
        threading.Thread.__init__(self)

    def run(self):
        send_email(sender=self.sender, subject=self.subject, mail_content=self.mail_content,
                   recipient_list=self.recipient_list, template_name=self.template_name, cc_list=self.cc_list,
                   bcc_list=self.bcc_list, reply_to=self.reply_to, fail_silently=self.fail_silently)


def filter_email_for_non_prod_environment(email_address_list):
    return email_address_list


def append_non_prod_env_to_subject(subject):
    if settings.CURRENT_RUN_ENVIRONMENT is not settings.PROD_ENVIRONMENT:
        # append env to subject
        return "[" + settings.CURRENT_RUN_ENVIRONMENT + "] " + subject

    return subject


def send_email(
        sender,
        subject,
        mail_content,
        recipient_list,
        template_name=None,
        cc_list=None,
        bcc_list=None,
        reply_to=None,
        fail_silently=False):
    """
    send notification
    :param sender:
    :param subject:
    :param mail_content:
    :param recipient_list:
    :param template_name:
    :param cc_list:
    :param bcc_list:
    :param reply_to:
    :param fail_silently: used for testing
    """
    if template_name:
        template = get_template(template_name)
        ctx = Context({'request': mail_content})
        message = template.render(ctx)
    else:
        message = mail_content

    sender_email = sender
    if not isinstance(sender, str):
        if isinstance(sender, Iterable):
            for s in sender:
                sender_email = s
                break

    subject = append_non_prod_env_to_subject(subject)

    to_list = filter_email_for_non_prod_environment(recipient_list)
    email = EmailMessage(subject=subject, body=message,
                         from_email=sender_email, to=to_list)
    if cc_list:
        email.cc = filter_email_for_non_prod_environment(cc_list)
    if bcc_list:
        email.bcc = filter_email_for_non_prod_environment(bcc_list)
    if reply_to:
        email.reply_to = filter_email_for_non_prod_environment(reply_to)

    email.send(fail_silently=fail_silently)


def send_notification(sender, subject, mail_content, recipient_list, template_name=None, cc_list=None, bcc_list=None,
                      reply_to=None, fail_silently=False):
    EmailThread(sender, subject, mail_content, recipient_list, template_name, cc_list, bcc_list, reply_to,
                fail_silently).start()



# Generated by Django 3.1.1 on 2020-11-24 23:17

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('authtoken', '0002_auto_20160226_1747'),
        ('crams', '0005_support'),
    ]

    operations = [
        migrations.CreateModel(
            name='CramsToken',
            fields=[
                ('token_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='authtoken.token')),
                ('ks_roles', models.TextField(blank=True, null=True)),
            ],
            bases=('authtoken.token',),
        ),
    ]



# coding=utf-8
"""
    Utilities to manipulate django model fields
"""
from django.core import serializers
from rest_framework import mixins, viewsets
from rest_framework import response
from rest_framework import status

# noinspection PyProtectedMember
from crams.constants import db
from django.conf import settings


class CramsModelNoUpdateViewSet(mixins.CreateModelMixin,
                                mixins.RetrieveModelMixin,
                                mixins.ListModelMixin,
                                viewsets.GenericViewSet):
    """
        A viewset similar to viewset.ModelViewSet except
        - no destroy() action
        - partial update will return error message
    """

    def partial_update(self, request, *args, **kwargs):
        msg = {
            "detail": "Method \"PATCH\" not allowed."
        }
        headers = self.get_success_headers(msg)
        return response.Response(msg,
                                 status=status.HTTP_405_METHOD_NOT_ALLOWED,
                                 headers=headers)


class CramsModelViewSet(CramsModelNoUpdateViewSet, mixins.UpdateModelMixin):
    pass


class CramsModelNoListViewSet(CramsModelViewSet):
    def list(self, request, *args, **kwargs):
        msg = {
            "detail": "Method \"GET\" not allowed."
        }
        return response.Response(msg,
                                 status=status.HTTP_405_METHOD_NOT_ALLOWED)


def list_model_fields(model_instance):
    """
        List model fields for given instance
    :param model_instance:
    """
    for f in model_instance._meta.get_all_field_names():
        print(repr(f))


# noinspection PyProtectedMember
def list_model_field_values(model_instance):
    # modify : for django 1.9 use model._meta.get_fields()  and  field.name
    """
        List model field and values for given instance
    :param model_instance:
    """
    for f in model_instance._meta.get_all_field_names():
        print(f, getattr(model_instance, f))


def get_model_field_value(model_instance, field_str):
    """
        List value for a given model field
    :param model_instance:
    :param field_str:
    :return:
    """
    return getattr(model_instance, field_str)


def json_serialize(query_set):
    """
        serialize in query_set as json
    :param query_set:
    :return:
    """
    return serializers.serialize("json", query_set)


def get_referer_host_url(request):
    base_url = None
    if request:
        base_url = request.META.get('HTTP_REFERER')
    return base_url


def generate_client_login_url(request, base_url):
    if not base_url:
        base_url = get_referer_host_url(request)

    return base_url + settings.CLIENT_KS_LOGIN_PATH


def get_funding_body_base_url(funding_body_name):
    return FUNDING_BODY_BASE_URL_PATH.get(funding_body_name.lower(), '')


def get_funding_body_join_url(funding_body_name):
    return FUNDING_BODY_JOIN_PATH.get(funding_body_name.lower(), '')


def get_funding_body_member_url(funding_body_name):
    return FUNDING_BODY_MEMBER_PATH.get(funding_body_name.lower(), '')


def get_funding_body_request_url(funding_body_name):
    return FUNDING_BODY_CLIENT_REQUEST_PATH.get(funding_body_name.lower(), '')


def get_funding_body_approval_url(funding_body_name):
    return FUNDING_BODY_CLIENT_APPROVAL_PATH.get(funding_body_name.lower(), '')


FUNDING_BODY_BASE_URL_PATH = {

}

FUNDING_BODY_JOIN_PATH = {

}

FUNDING_BODY_MEMBER_PATH = {

}

FUNDING_BODY_CLIENT_REQUEST_PATH = {

}

FUNDING_BODY_CLIENT_APPROVAL_PATH = {

}



# coding=utf-8
"""
 methods to fetch Lookup Data
"""
from rest_framework.exceptions import ParseError

from crams.models import EResearchBodySystem, EResearchBodyIDKey
from crams.models import FundingScheme, FundingBody
from crams.models import Provider
from crams.models import Question
from crams.models import FORCode


class LookupDataModel:

    """

    :param model:
    """

    def __init__(self, model):
        self.model = model

    def get_lookup_data(self, search_key_dict):
        """

        :param search_key_dict:
        :return: :raise ParseError:
        """
        try:
            return self.model.objects.get(**search_key_dict)
        except self.model.DoesNotExist:
            raise ParseError(self.model.__name__ +
                             ' does not exist for search_key_dict ' +
                             repr(search_key_dict))
        except self.model.MultipleObjectsReturned:
            raise ParseError(self.model.__name__ +
                             ' Multiple objects found for search_key_dict ' +
                             repr(search_key_dict))

    def serialize(self, search_key_dict, serializer):
        """

        :param search_key_dict:
        :param serializer:
        :return: :raise ParseError:
        """
        try:
            return serializer(self.get_lookup_data(search_key_dict)).data
        except Exception:
            raise ParseError(
                self.model.__name__ + ': error serializing data for search' +
                                      '_key_dict ' + repr(search_key_dict))


def get_e_research_system_obj(search_key_dict):
    """

    :param search_key_dict:
    :return:
    """
    return LookupDataModel(EResearchBodySystem).get_lookup_data(search_key_dict)


def get_question_obj(search_key_dict):
    """

    :param search_key_dict:
    :return:
    """
    return LookupDataModel(Question).get_lookup_data(search_key_dict)


def get_provider_obj(search_key_dict):
    """

    :param search_key_dict:
    :return:
    """
    return LookupDataModel(Provider).get_lookup_data(search_key_dict)


def get_system_obj(search_key_dict):
    """

    :param search_key_dict:
    :return:
    """
    return LookupDataModel(EResearchBodyIDKey).get_lookup_data(search_key_dict)


def get_for_code_obj(search_key_dict):
    """

    :param search_key_dict:
    :return:
    """
    return LookupDataModel(FORCode).get_lookup_data(search_key_dict)


def get_funding_scheme_obj(search_key_dict):
    """

    :param search_key_dict:
    :return:
    """
    return LookupDataModel(FundingScheme).get_lookup_data(search_key_dict)


def get_funding_body_obj(search_key_dict):
    """

    :param search_key_dict:
    :return:
    """
    return LookupDataModel(FundingBody).get_lookup_data(search_key_dict)



# coding=utf-8
"""

"""
# from django.db.models import Q
# from rest_framework import exceptions
#
# from crams.config import config_init
# from crams import models
# from crams.utils.role import AbstractCramsRoleUtils as roleUtils
# from crams.manager import models as manager_models


# def fetch_user_contact_if_exists(crams_user):
#     return models.Contact.fetch_contact_for_user(crams_user)
#
# def user_details_json(user_obj, null_default='System'):
#     if user_obj:
#         contact = fetch_user_contact_if_exists(user_obj)
#         if contact:
#             return {
#                 'email': contact.email,
#                 'first_name': contact.given_name,
#                 'last_name': contact.surname
#             }
#         return {
#             'email': user_obj.email,
#             'first_name': user_obj.first_name,
#             'last_name': user_obj.last_name
#         }
#     return {
#         'email': null_default,
#         'first_name': null_default,
#         'last_name': null_default
#     }
#
#
# def fetch_related_user_erb_system_list(serializer_self):
#     user_obj, _ = get_current_user_from_context(serializer_self)
#     if user_obj:
#         return roleUtils.get_authorised_e_research_system_list(user_obj)
#     return list()
#
#
# def fetch_related_user_erb_list(serializer_self):
#     ret_set = set()
#     for erbs in fetch_related_user_erb_system_list(serializer_self):
#         ret_set.add(erbs.e_research_body)
#     return list(ret_set)
#
#
# def fetch_erb_userroles_with_provision_privileges(
#    # moved to crams_contact.utils.contact_utils
#
#
# def fetch_contact_delegates_for_erb_systems(contact, erb_systems):
#     erb_set = set()
#     for erbs in erb_systems:
#         erb_set.add(erbs.e_research_body)
#     role_qs = contact.user_roles.filter(role_erb__in=erb_set,
#                                         delegates__isnull=False)
#     ret_set = set()
#     if role_qs.exists():
#         for erb_role in role_qs.all():
#             ret_set = ret_set | set(erb_role.delegates.all())
#     return list(ret_set)
#
#
# def fetch_contact_organisation_qs(contact):
#     org_pk_set = set()
#     for org_manager in contact.manager_organisations.filter(active=True):
#         org_pk_set.add(org_manager.organisation.id)
#     return models.Organisation.objects.filter(pk__in=org_pk_set)
#
#
# def fetch_contact_faculties_qs(contact):
#     faculty_pk_set = set()
#     for erb_user_role in contact.manager_faculties.filter(active=True):
#         faculty_pk_set.add(erb_user_role.faculty.id)
#
#     org_qs = fetch_contact_organisation_qs(contact)
#
#     qs_filter = Q(organisation__in=org_qs) | Q(pk__in=faculty_pk_set)
#     return models.Faculty.objects.filter(qs_filter)


# def fetch_contact_departments_qs(contact):
#     pk_set = set()
#     qs_all = manager_models.DepartmentManager.objects
#     for department_manager in qs_all.filter(active=True, contact=contact):
#         pk_set.add(department_manager.department.id)
#
#     faculty_qs = fetch_contact_faculties_qs(contact)
#
#     qs_filter = Q(faculty__in=faculty_qs) | Q(pk__in=pk_set)
#     return models.Department.objects.filter(qs_filter)
#
#
# def get_user_projects(user_obj):
#     qs_filter = Q(project_contacts__contact__email__iexact=user_obj.email)
#     return models.Project.objects.filter(qs_filter)
#
#
# def get_user_storage_requests(
#         user_obj, storage_product_qs=None, request_status_code=None):
#     user_projects = get_user_projects(user_obj)
#     storage_request_qs = models.StorageRequest.objects.filter(
#         request__project__in=user_projects)
#
#     if storage_product_qs:
#         storage_request_qs = storage_request_qs.filter(
#             storage_product__in=storage_product_qs)
#
#     if request_status_code:
#         return storage_request_qs.filter(
#             request__request_status__code=request_status_code)
#
#     return storage_request_qs
#
#
# def get_contact_historical_projects_qs(contact_obj):
#     # Do not use distinct here, instead use it in calling method
#     qs = models.Project.objects.filter(
#         project__project_contacts__contact=contact_obj)
#     return qs
#
#
# def get_contact_current_projects_qs(contact_obj):
#     # Do not use distinct here, instead use it in calling method
#     qs = get_contact_historical_projects_qs(contact_obj).filter(
#         current_project__isnull=True)
#     return qs
#
#
# def get_contact_storage_requests(
#         contact_obj, storage_product_qs=None,
#         request_status_code_list=None, ignore_status_code_for_current=False):
#
#     ret_qs = models.StorageRequest.objects.none()
#
#     qs_filter = Q()
#     if storage_product_qs:
#         if not storage_product_qs.exists():
#             return ret_qs
#         qs_filter = Q(storage_product__in=storage_product_qs)
#
#     if request_status_code_list:
#         request_status_qs = models.RequestStatus.objects.filter(
#             code__in=request_status_code_list)
#         if not request_status_qs.exists():
#             return ret_qs
#         status_filter = Q(request__request_status__in=request_status_qs)
#         if ignore_status_code_for_current:
#             status_filter = status_filter | Q(
#                 request__current_request__isnull=True)
#         qs_filter = qs_filter & status_filter
#
#     contact_projects = get_contact_current_projects_qs(contact_obj)
#     if not contact_projects.exists():
#         return ret_qs
#
#     project_filter = Q(request__project__in=contact_projects) | Q(
#         request__project__current_project__in=contact_projects)
#
#     qs_filter = qs_filter & project_filter
#     storage_request_qs = models.StorageRequest.objects.filter(qs_filter)
#
#     return storage_request_qs
#
#
# def is_contact_product_provider(contact, product):
#     """
#     return True is contact is a provider for the given product, else False
#     :param contact:
#     :param product:
#     :return:
#     """
#     fn_msg = 'user_utils.is_contact_product_provider: '
#     if not isinstance(contact, models.Contact):
#         raise exceptions.ValidationError(fn_msg + 'contact object required')
#     if not isinstance(product, models.ProductCommon):
#         raise exceptions.ValidationError(fn_msg + 'product object required')
#
#     product_erb = product.e_research_system.e_research_body
#     erb_role_qs = contact.user_roles.filter(role_erb=product_erb)
#     if erb_role_qs.exists():
#         erb_role = erb_role_qs.first()
#         # Check if Provider role exists
#         if product.provider and erb_role.providers.exists():
#             if product.provider in erb_role.providers.all():
#                 return True
#
#         # Check if erbAdmin and not in disallow Provider role list
#         if product_erb not in config_init.ERB_PROVIDER_DISALLOW_LIST:
#             if erb_role.is_erb_admin:
#                 return True
#
#         # Check if erb System Admin and not in disallow Provider role list
#         product_erbs = product.e_research_system
#         if product_erbs not in config_init.ERBS_PROVIDER_DISALLOW_LIST:
#             if product_erbs in erb_role.admin_erb_systems.all():
#                 return True
#     return False
#
#
# def is_project_visible_to_user(user_obj):
#     contact = fetch_user_contact_if_exists(crams_user=user_obj)
#     if not contact:
#         return False
#
#
# def fetch_project_contact_via_role(prj_contacts, role_str):
#     contact_list = []
#     for pc in prj_contacts:
#         if pc.contact_role.name.lower() == role_str.lower():
#             contact_list.append(pc.contact)
#
#     return contact_list



# Generated by Django 3.2 on 2021-07-06 06:32

from django.db import migrations, models

class Migration(migrations.Migration):

    dependencies = [
        ('crams', '0008_load_crams_lookups'),
    ]

    operations = [
        migrations.CreateModel(
            name='ProjectMemberStatus',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('code', models.CharField(max_length=50)),
                ('status', models.CharField(max_length=100)),
            ],
        ),
    ]





# coding=utf-8
"""
 validation utils
"""
from abc import ABCMeta

from rest_framework.serializers import ValidationError
from crams.utils.django_utils import get_model_field_value


class ChangedMetadata:
    def __init__(self):
        self.obj_dict = dict()
        self.modified_related_field_set = set()
        self.new_related_field_set = set()
        self.removed_related_field_set = set()
        self.err_list = list()


class FieldsRequiredValidator(metaclass=ABCMeta):
    """class FieldsRequiredValidator."""

    @classmethod
    def get_fields_required(cls):
        """get required fields."""
        return list()

    @classmethod
    def get_fields_unchanged(cls):
        """get fields unchanged"""
        pass

    def __call__(self, obj_dict):
        for field in self.get_fields_required():
            if field not in obj_dict:
                raise ValidationError({field: 'field is required'})


def validate_dry_principle(serializer_instance, fields_unchanged_list,
                           list_name_str, user_role_str):
    """validate dry principle.

    :param serializer_instance:
    :param fields_unchanged_list:
    :param list_name_str:
    :param user_role_str:
    :return: :raise ValidationError:
    """
    if not list_name_str:
        raise ValidationError(
            {'listNameStr / validate_DRY ': 'parameter cannot be null, ' +
                                            'contact Tech Support'})

    if not serializer_instance.instance:
        raise ValidationError({'{}'.format(
            repr(serializer_instance)): 'instance not found, contact ' +
                                        'Tech Support'})

    # Do not use the parsed validatedData, it does not contain id pk
    intial_data_list = serializer_instance.initial_data.get(list_name_str,
                                                            None)
    existing_instance_list = get_model_field_value(
        serializer_instance.instance, list_name_str).all()

    msg = None
    if not intial_data_list:
        if existing_instance_list:
            msg = '{} cannot remove existing {}'.format(
                user_role_str, list_name_str)
    elif not existing_instance_list:
        msg = '{} cannot add new {}'.format(user_role_str, list_name_str)
    elif len(intial_data_list) < len(existing_instance_list):
        msg = '{} cannot remove existing {}'.format(user_role_str,
                                                    list_name_str)
    elif len(intial_data_list) > len(existing_instance_list):
        msg = '{} cannot add new {}'.format(user_role_str,
                                            list_name_str)

    if msg:
        raise ValidationError({list_name_str: msg})

    existing_instance_map = {}
    for cr in existing_instance_list:
        existing_instance_map[cr.id] = cr

    def verify_fields():  # useful inner function
        """verify fields.

        :return: :raise ValidationError:
        """

        def recursive_verify(field_list, existing, in_data, parent_field_str):
            """recursive verify.

            :param field_list:
            :param existing:
            :param in_data:
            :param parent_field_str:
            :raise ValidationError:
            """
            for field in field_list:
                if isinstance(field, dict):
                    for k in field.keys():
                        model_instance = get_model_field_value(existing, k)
                        if not model_instance:
                            raise ValidationError(
                                {parent_field_str: '{} field not found in ' +
                                    'model {}'.format(k, type(existing))})
                        in_data_value = in_data.get(k, None)
                        if not in_data_value:
                            raise ValidationError(
                                {parent_field_str: '{} field required'.
                                    format(k)})
                        recursive_verify(
                            field.get(k),
                            model_instance,
                            in_data_value,
                            '{} {} (id: {})'.format(
                                parent_field_str,
                                k,
                                repr(
                                    model_instance.id)))
                else:
                    in_data_value = in_data.get(field, None)
                    if in_data_value:
                        if get_model_field_value(existing, field) \
                                != in_data_value:
                            # noinspection PyPep8
                            raise ValidationError(
                                {parent_field_str: '{} ' + 'value cannot be \
                                 changed by {} to {}'.format(field,
                                 user_role_str, repr(in_data_value))})

        return recursive_verify

    for init_data in intial_data_list:
        in_id = init_data.pop('id', None)
        if not in_id:
            raise ValidationError({list_name_str: 'id is required'})

        existing_instance = existing_instance_map.pop(in_id, None)
        if not existing_instance:
            raise ValidationError(
                {list_name_str: 'missing object with id {}'.
                    format(repr(in_id))})

        validator = verify_fields()
        validator(fields_unchanged_list, existing_instance, init_data,
                  '{} (id: {} )'.format(list_name_str, repr(in_id)))



# coding=utf-8
"""
Role Utilities
"""
import abc
from json import dumps as json_dumps
from json import loads as json_loads
from rest_framework import exceptions

from crams import models
from crams.utils import lang_utils


class AbstractCramsRoleUtils(abc.ABC):
    """
    It is expected that the CramsToken.ks_roles field will be a json dict
      - with the following keys whose values are a list of roles
    The abstract method build_user_roles below should return this dict
    """
    ADMIN_ROLE_KEY = 'admin'
    APPROVER_ROLE_KEY = 'approver'
    PROVIDER_ROLE_KEY = 'provisioner'
    DELEGATE_ROLE_TYPE = 'delegate'
    TENANT_MANAGER = 'tenant_manager'
    SERVICE_MANAGEMENT_ROLE_TYPE = 'service_management'
    FACULTY_MANAGEMENT_ROLE_TYPE = 'faculty_management'
    LEGACY_ROLES_KEY = 'global'

    @abc.abstractmethod
    def build_user_roles(self, user_obj):
        """
        return a dict whose keys are the role KEYs defined above
          - whose values are list of roles
        """
        msg = 'abstract method build_user_roles not implemented'
        raise exceptions.ParseError(msg)

    @classmethod
    def generate_project_role(cls, project_name, role_name):
        return project_name + '_' + role_name

    @classmethod
    def fetch_cramstoken_roles_dict(cls, userobj):
        """
        Fetch roles stored in the DB for given user object
        :param userobj:
        :return:
        """
        if not userobj:
            return None

        # Token could have changed since the user object was fetched.
        #   get latest user object to read latest token
        try:
            current_userobj = models.User.objects.get(pk=userobj.id)
            if current_userobj and hasattr(current_userobj, 'auth_token'):
                auth_token = current_userobj.auth_token
                if hasattr(auth_token, 'cramstoken') and \
                        auth_token.cramstoken.ks_roles:
                    return json_loads(auth_token.cramstoken.ks_roles)
        except models.User.DoesNotExist:
            pass
        return None

    @classmethod
    def get_authorised_e_research_system_list(cls, user, debug=False):
        """

        :param user:
        :return:
        """
        ret_set = set()
        role_dict = cls.fetch_cramstoken_roles_dict(user)
        if debug:
            print('role_dict fetched', role_dict)

        if role_dict:
            if cls.SERVICE_MANAGEMENT_ROLE_TYPE in role_dict:
                return list(models.EResearchBodySystem.objects.all())

            admin_roles = role_dict.get(cls.ADMIN_ROLE_KEY)
            approver_roles = role_dict.get(cls.APPROVER_ROLE_KEY)

            for erbsystem in models.EResearchBodySystem.objects.all():
                if debug:
                    print( '     -- checking for role',
                           erbsystem.admin_role, erbsystem.e_research_body.admin_role, erbsystem.approver_role)
                if admin_roles:
                    if erbsystem.admin_role in admin_roles \
                            or erbsystem.e_research_body.admin_role in admin_roles:
                        ret_set.add(erbsystem)
                if approver_roles and erbsystem.approver_role in approver_roles:
                    ret_set.add(erbsystem)

            if approver_roles:
                for delegate in models.EResearchBodyDelegate.objects.all():
                    if delegate.approver_role in approver_roles:
                        for erbsystem in delegate.e_research_body.systems.all():
                            ret_set.add(erbsystem)

        return list(ret_set)

    @classmethod
    def get_authorised_provider_list(cls, user):
        """

        :param user:
        :return:
        """
        ret_list = list()
        user_roles_dict = cls.fetch_cramstoken_roles_dict(user)
        if user_roles_dict and cls.PROVIDER_ROLE_KEY in user_roles_dict:
            provider_roles = user_roles_dict.get(cls.PROVIDER_ROLE_KEY)
            for p in models.Provider.objects.all():
                if p.provider_role in provider_roles:
                    ret_list.append(p)
        return ret_list

    def setup_crams_token_and_roles(self, user):
        crams_token, created = models.CramsToken.objects.get_or_create(user=user)
        if crams_token.is_expired():  # Expire existing Token
            crams_token.delete()
            crams_token = models.CramsToken(user=user)

        crams_token.ks_roles = json_dumps(self.build_user_roles(user))
        crams_token.save()
        return crams_token

    @classmethod
    def fetch_cramstoken_roles_flatten_to_list(cls, userobj):
        ret_roles_list = list()

        user_roles_dict = cls.fetch_cramstoken_roles_dict(userobj)
        if user_roles_dict:
            for key, roles in user_roles_dict.items():
                if roles:
                    ret_roles_list.append(key)
                    ret_roles_list.extend(roles[:])
        return ret_roles_list

    @classmethod
    def has_e_system_role(cls, user_obj, erb_system_obj=None):
        """

        :param user_obj:
        :param erb_system_obj:
        :return:
        """
        user_roles_dict = cls.fetch_cramstoken_roles_dict(user_obj)
        if user_roles_dict:
            if cls.APPROVER_ROLE_KEY in user_roles_dict:
                if not erb_system_obj:
                    return True
                approver_roles = user_roles_dict.get(cls.APPROVER_ROLE_KEY)
                required_role = erb_system_obj.approver_role
                if approver_roles and required_role in approver_roles:
                    return True

            if cls.ADMIN_ROLE_KEY not in user_roles_dict:
                if not erb_system_obj:
                    return True
                admin_roles = user_roles_dict.get(cls.ADMIN_ROLE_KEY)
                if admin_roles and erb_system_obj.admin_role in admin_roles:
                    return True

        return False

    @classmethod
    def user_has_roles(cls, userobj, role_list):
        role_set = set([lang_utils.strip_lower(role) for role in role_list])
        user_roles_list = cls.fetch_cramstoken_roles_flatten_to_list(userobj)
        if user_roles_list:
            return role_set.issubset(set(user_roles_list))
        return False

    @classmethod
    def user_has_atleast_one_role(cls, userobj, role_list):
        role_set = set([lang_utils.strip_lower(role) for role in role_list])
        user_roles_dict = cls.fetch_cramstoken_roles_dict(userobj)
        if user_roles_dict:
            for key, role in user_roles_dict.items():
                if role in role_set:
                    return True
        return False

    @classmethod
    def is_user_erb_admin(cls, user_obj, erb_obj=None,
                          erb_system_obj=None, include_systems=False):
        user_roles_dict = cls.fetch_cramstoken_roles_dict(user_obj)
        if not user_roles_dict or cls.ADMIN_ROLE_KEY not in user_roles_dict:
            return False
        if erb_obj:
            if not isinstance(erb_obj, models.EResearchBody):
                return False
            admin_role_list = user_roles_dict.get(cls.ADMIN_ROLE_KEY)
            if admin_role_list and erb_obj.admin_role in admin_role_list:
                return True
            if include_systems:
                for erb_system in erb_obj.systems.all():
                    if erb_system.admin_role in admin_role_list:
                        return True
        if erb_system_obj:
            if not isinstance(erb_system_obj, models.EResearchBodySystem):
                return False
            admin_role_list = user_roles_dict.get(cls.ADMIN_ROLE_KEY)
            return admin_role_list and erb_system_obj.admin_role in admin_role_list
        return True

    @classmethod
    def is_user_a_provider(cls, user_obj):
        user_roles_dict = cls.fetch_cramstoken_roles_dict(user_obj)
        return user_roles_dict and cls.PROVIDER_ROLE_KEY in user_roles_dict



# Generated by Django 3.1.1 on 2020-09-25 02:30

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('crams', '0003_lookup'),
    ]

    operations = [
        migrations.CreateModel(
            name='ProvisionDetails',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('creation_ts', models.DateTimeField(auto_now_add=True)),
                ('last_modified_ts', models.DateTimeField(auto_now=True)),
                ('status', models.CharField(choices=[('S', 'Sent'), ('P', 'Provisioned'), ('F', 'Failed'), ('L', 'Resend'), ('U', 'Updated'), ('X', 'Update Sent'), ('?', '???')], default='S', max_length=1)),
                ('message', models.TextField(blank=True, null=True)),
                ('created_by', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='provisiondetails_created_by', to=settings.AUTH_USER_MODEL)),
                ('parent_provision_details', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='history', to='crams.provisiondetails')),
                ('updated_by', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='provisiondetails_updated_by', to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='Provider',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('creation_ts', models.DateTimeField(auto_now_add=True)),
                ('last_modified_ts', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=200)),
                ('start_date', models.DateField(auto_now_add=True)),
                ('active', models.BooleanField(default=True)),
                ('description', models.TextField(blank=True, null=True)),
                ('created_by', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='provider_created_by', to=settings.AUTH_USER_MODEL)),
                ('updated_by', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='provider_updated_by', to=settings.AUTH_USER_MODEL)),
            ],
        ),
    ]



# coding=utf-8
"""

"""
from django.db import models
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _
from django.contrib.auth import models as auth_models


class UserManager(auth_models.BaseUserManager):
    def create_user(self, email, username, password=None):
        """
        Creates and saves a User with the given email, date of
        birth and password.
        """
        if not email:
            raise ValueError('Users must have an email address')

        user = self.model(
            email=email,
            username=username,
        )

        user.set_password(password)
        user.save(using=self._db)
        return user

    def create_superuser(self, email, username, password):
        """
        Creates and saves a superuser with the given email, date of
        birth and password.
        """
        user = self.create_user(email, password=password, username=username)
        user.is_admin = True
        user.save(using=self._db)
        return user



# coding=utf-8
"""
Crams Permissions
"""
import abc
from rest_framework import permissions, exceptions

from django.conf import settings
from crams.models import CramsToken
from crams.utils.role import AbstractCramsRoleUtils


class IsCramsAuthenticated(permissions.IsAuthenticated):
    """
    Global permission, determine if Curent User has provider role
    """
    message = 'User does not hold valid CramsToken.'

    def has_permission(self, request, view):
        """
        user has a valid cramstoken (not expired)
        :param request:
        :param view:
        :return:
        """
        if super().has_permission(request, view):
            try:
                crams_token = CramsToken.objects.get(user=request.user)
                if crams_token.is_expired():
                    raise exceptions.NotAuthenticated(self.message)
                return True
            except Exception as e:
                pass
        return False


class IsCramsTestSetupUser(IsCramsAuthenticated):
    """
    Global permission, determine if Curent User has provider role
    """
    message = 'User is not authorized to setup test users, ' \
              'update settings.VALID_TEST_SETUP_USER_EMAIL_LIST and check crams token expiry'

    def has_permission(self, request, view):
        """
        user has a valid cramstoken (not expired)
        :param request:
        :param view:
        :return:
        """
        print('in IsCramsTestSetupUser')
        if settings.CURRENT_RUN_ENVIRONMENT.lower() in ['qat', 'staging']:
            if super().has_permission(request, view):
                print('valid emails', settings.VALID_TEST_SETUP_USER_EMAIL_LIST)
                return request.user.email.lower() in settings.VALID_TEST_SETUP_USER_EMAIL_LIST
            else:
                print('User auth failed, either token expired or wrong auth')

        return False


class IsLookupAdmin(permissions.BasePermission):
    """
    Object-level permission to allow Faculty level admin tasks
    """
    message = 'User is not a Faculty Admin'

    def has_object_permission(self, request, view, obj):
        """
        User is assigned as admin to given faculty
        :param request:
        :param view:
        :param faculty obj:
        :return:
        """
        if request.method in permissions.SAFE_METHODS:
            return True

        user_roles_dict = AbstractCramsRoleUtils.fetch_cramstoken_roles_dict(request.user)
        if user_roles_dict:
            return AbstractCramsRoleUtils.ADMIN_ROLE_KEY in user_roles_dict
        return False


class IsCramsAdmin(permissions.BasePermission):
    """
    Global permission to allow CRAMS admin
    """
    message = 'User does not hold Crams Admin role.'

    def has_permission(self, request, view):
        """
        user has CRAMS_ADMIN_ROLE role
        :param request:
        :param view:
        :return:
        """

        user_roles_dict = AbstractCramsRoleUtils.fetch_cramstoken_roles_dict(request.user)
        if user_roles_dict:
            return AbstractCramsRoleUtils.ADMIN_ROLE_KEY in user_roles_dict
        return False

    def has_object_permission(self, request, view, obj):
        return True


class IsServiceManager(permissions.BasePermission):
    """
    Global permission to allow Service Manager role
    """
    message = 'User does not hold Service Manager role.'

    @classmethod
    def role_exists_for_user(cls, user):
        user_roles_dict = AbstractCramsRoleUtils.fetch_cramstoken_roles_dict(user)
        if user_roles_dict:
            return AbstractCramsRoleUtils.SERVICE_MANAGEMENT_ROLE_TYPE in user_roles_dict
        return False

    def has_permission(self, request, view):
        """
        user has Service Manager role
        :param request:
        :param view:
        :return:
        """
        return self.role_exists_for_user(request.user)


class IsAbstractFacultyManager(permissions.BasePermission):
    """
    Global permission to allow faculty Manager role
    """
    message = 'User does not hold Faculty Manager role.'

    @classmethod
    def role_exists_for_user(cls, user_obj):
        exists, _ = cls.user_roles_common(user_obj)
        return exists

    @classmethod
    def user_roles_common(cls, user_obj):
        user_roles_dict = AbstractCramsRoleUtils.fetch_cramstoken_roles_dict(user_obj)

        role_exists = False
        if user_roles_dict:
            role_exists = \
                AbstractCramsRoleUtils.FACULTY_MANAGEMENT_ROLE_TYPE in user_roles_dict
        return role_exists, user_roles_dict

    def has_permission(self, request, view):
        """
        user has Faculty Manager role
        :param self:
        :param request:
        :param view:
        :return:
        """

        role_exists, user_roles_dict = self.user_roles_common(request.user)
        return role_exists

    @abc.abstractmethod
    def has_access_to_obj(self, user, obj):
        msg = 'abstract method has_access_to_obj not implemented'
        raise exceptions.ParseError(msg)

    def has_object_permission(self, request, view, obj):
        return self.has_access_to_obj(request.user, obj)


class IsActiveProvider(permissions.BasePermission):
    """
    Global permission, determine if Curent User has provider role
    """
    message = 'User does not hold Crams Provisioner role.'

    def has_permission(self, request, view):
        """
        user has CRAMS_PROVISIONER_ROLE role
        :param request:
        :param view:
        :return:
        """
        user_roles_dict = AbstractCramsRoleUtils.fetch_cramstoken_roles_dict(request.user)
        if user_roles_dict:
            return AbstractCramsRoleUtils.PROVIDER_ROLE_KEY in user_roles_dict



# coding=utf-8
"""
    Validate Config
"""
import logging
import importlib
from django.conf import settings
from crams.utils import role
from crams.constants import db

MINIMUM = 'minimum'
MAXIMUM = 'maximum'

# Research Body Reply-to Email Map
eSYSTEM_REPLY_TO_EMAIL_MAP = dict()

# Disallow admin Provider role for the following ERB Systems
ERBS_PROVIDER_DISALLOW_LIST = list()
# Disallow admin Provider role for the following Eresearch Body
ERB_PROVIDER_DISALLOW_LIST = list()

# Project ErbId related dict
#     key: (id_key_name, erb_name)
#     value: (generator_fn, validate_fn)
ERBS_PROJECT_ID_UPDATE_FN_MAP = dict()

# Contact ErbId related
ERBS_CONTACT_ID_UPDATE_FN_MAP = dict()

# EResearchSystem Delegate Question Key Map
DELEGATE_QUESTION_KEY_MAP = dict()

# Contact System Identifier Email Template
CONTACT_SYSTEM_IDENTIFIER_EMAIL_TEMPLATE = dict()

# Update User Visibility
CRAMS_DEFAULT_UPDATED_BY_VISIBILTY_ROLE = [role.AbstractCramsRoleUtils.APPROVER_ROLE_KEY,
                                           role.AbstractCramsRoleUtils.PROVIDER_ROLE_KEY]
DEFAULT_UPDATED_BY_PREFIX = ''  # 'Authorised User'
UPDATED_BY_VISIBLE_ROLEGRP_FOR_ERB_LOWER = dict()

# Allocation list API requires returning of ERB system specific Project Ids like HPC_POSIX_GROUP_ID
ALLOCATION_LIST_ERB_IDKEY_DICT = dict()

# Each Role can be set to occur a minimum and maximum number of times.
# A value of 0 in max implies unlimited
CONTACT_ROLE_VALID_COUNT = dict()
CONTACT_ROLE_VALID_COUNT[db.eSYSTEM_NECTAR] = {db.CHIEF_INVESTIGATOR: (1, 1)}
CONTACT_ROLE_VALID_COUNT[db.eSYSTEM_VICNODE] = {db.DATA_CUSTODIAN: (1, 1),
                                                db.TECHNICAL_CONTACT: (1, 0)}

# for config_settings in settings.CRAMS_APP_CONFIG_LIST:
#     try:
#         importlib.import_module(config_settings)  # noqa
#     except ImportError:
#         logging.debug("config file not found: {}".format(config_settings))


def get_min_max_dict(min=None, max=None):
    ret_dict = dict()
    if min:
        ret_dict[MINIMUM] = min
    if max:
        ret_dict[MAXIMUM] = max
    return ret_dict



# coding=utf-8
"""

"""

from rest_framework.renderers import BrowsableAPIRenderer


def get_current_user_from_context(serializer_self):
    current_user = None
    context = None

    if hasattr(serializer_self, 'context'):
        context = serializer_self.context
        current_user = context.get('current_user', None)
        if not current_user:
            http_request = serializer_self.context.get('request', None)
            if http_request:
                current_user = http_request.user
    elif serializer_self.request:
        current_user = serializer_self.request.user
        context = dict()
        context['request'] = serializer_self.request

    return current_user, context


# reference: https://bradmontgomery.net/blog/
#                disabling-forms-django-rest-frameworks-browsable-api/
class BrowsableAPIRendererWithoutForms(BrowsableAPIRenderer):
    """Renders the browsable api, but excludes the forms."""

    def get_context(self, *args, **kwargs):
        ctx = super().get_context(*args, **kwargs)
        ctx['display_edit_forms'] = False
        return ctx

    def show_form_for_method(self, view, method, request, obj):
        """We never want to do this! So just return False."""
        return False

    def get_rendered_html_form(self, data, view, method, request):
        """Why render _any_ forms at all. This method should return
        rendered HTML, so let's simply return an empty string.
        """
        return ""



# coding=utf-8
"""utilitySerializers."""
import ast

from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from rest_framework import serializers, exceptions
from rest_framework.relations import RelatedField
from rest_framework.exceptions import ParseError, ValidationError


class UpdatableSerializer(object):
    """class UpdatableSerializer."""

    def update(self, instance, validated_data):
        """update.

        :param instance:
        :param validated_data:
        :return instance:
        """
        for f in self.model._meta.get_all_field_names():
            setattr(instance, f, validated_data.get(f, getattr(instance, f)))
        instance.save()
        return instance


class CreateOnlyModelSerializer(serializers.ModelSerializer):
    """class ReadOnlyModelSerializer."""

    def update(self, instance, validated_data):
        """update.

        :param instance:
        :param validated_data:
        """
        raise ParseError('Update not allowed ')

    def __delete__(self, instance):
        raise exceptions.APIException('Delete not Allowed')


class UpdateOnlyModelSerializer(serializers.ModelSerializer):
    """class UpdateOnlyModelSerializer."""

    def create(self, validated_data):
        """create.

        :param validated_data:
        """
        raise ParseError('Create not allowed ')

    def __delete__(self, instance):
        raise exceptions.APIException('Delete not Allowed')


class ReadOnlyModelSerializer(UpdateOnlyModelSerializer):
    """class ReadOnlyModelSerializer."""

    def update(self, instance, validated_data):
        """update.

        :param instance:
        :param validated_data:
        """
        raise ParseError('Update not allowed ')


class PrimaryKeyLookupField(serializers.PrimaryKeyRelatedField):
    """class PrimaryKeyLookupField.
        - cannot be used for optional serializer fields
    """

    def __init__(self, *args, **kwargs):
        """init.

        :param *args:
        :param **kwargs:
        """
        # Don't pass the 'fields' arg up to the superclass
        # Grant._meta.get_all_field_names()
        self.fields = kwargs.pop('fields', ['id'])
        self.pkFields = kwargs.pop('keyFields', ['id'])
        self.model = None

        # make sure this is a model class, not a string
        model = kwargs.pop('model', None)
        if model:
            msg = '{} is not an instance of django Model'.format(model)
            if not isinstance(model(), models.Model):
                raise ParseError(msg)
            self.model = model

        # Instantiate the superclass normally
        super(PrimaryKeyLookupField, self).__init__(*args, **kwargs)

    def to_internal_value(self, data):
        """to internal value.

        :param data:
        :return data:
        """
        if len(self.pkFields) > 0:
            data_dict = data
            if isinstance(data_dict, str):
                data_dict = ast.literal_eval(data)
            if not isinstance(data_dict, dict):
                raise ParseError('{} is not a dict object', data)
            ret_dict = {}
            for k in self.pkFields:
                try:
                    ret_dict[k] = data_dict[k]
                except Exception:
                    raise ParseError(
                        '{}: Field {} required'.format(self.label, k))
            return ret_dict

        return data

    def to_representation(self, value):
        """to representation.

        :param value:
        :return ret_dict:
        """

        if self.pk_field is not None:
            return self.pk_field.to_representation(value.pk)

        try:
            if self.model:
                instance = self.model.objects.get(pk=value.pk)
            else:
                if isinstance(value, dict):
                    instance = self.get_queryset().get(**value)
                elif value.pk is not None:
                    instance = self.get_queryset().get(pk=value.pk)
                else:
                    return None

            ret_dict = {}
            if self.fields:
                for k in self.fields:
                    if isinstance(k, dict):
                        for name, path in k.items():
                            node = instance
                            for attr in path.split('.'):
                                node = getattr(node, attr)
                            ret_dict[name] = node
                    else:
                        ret_dict[k] = getattr(instance, k)
            return ret_dict

        except ObjectDoesNotExist:
            self.fail('does_not_exist', pk_value=value.pk)
        except (TypeError, ValueError):
            self.fail('incorrect_type', data_type=type(value.pk).__name__)


class DynamicFieldsBaseSerializer(serializers.BaseSerializer):
    """class DynamicFieldsBaseSerializer.

    A BaseSerializer that takes an additional `fields` argument that
    controls which fields should be displayed dynamically at invocation
    """

    def __init__(self, *args, **kwargs):
        """init.

        :param *args:
        :param **kwargs:
        """
        # Don't pass the 'fields' arg up to the superclass
        self.displayFields = kwargs.pop('display', None)
        self.validateRequired = kwargs.pop('required', None)
        # Instantiate the superclass normally
        super(DynamicFieldsBaseSerializer, self).__init__(*args, **kwargs)

    def to_representation(self, obj):
        """to representation.

        :param obj:
        :param ret_dict:
        """
        ret_dict = {}
        if self.displayFields:
            for k in self.displayFields:
                ret_dict[k] = getattr(obj, k)

        return ret_dict

    def to_internal_value(self, data):
        """to internal value.

        :param data:
        :param ret_dict:
        """
        ret_dict = {}
        required = self.validateRequired
        if not required:
            required = ['id']
        for k in required:
            # Perform the data validation.
            val = data.get(k, None)
            if not val:
                raise ValidationError({
                    k: 'This field is required.'
                })
            ret_dict[k] = val

        return ret_dict


# refactored From http://tinyurl.com/ohsgt4g
class DynamicLookupModelSerializer(ReadOnlyModelSerializer):
    """class DynamicLookupModelSerializer.

    A ModelSerializer that takes an additional `fields` argument that
    controls which fields should be displayed.
    """

    class Meta(object):
        pass

    def __init__(self, *args, **kwargs):
        """init.

        :param *args:
        :param **kwargs:
        """
        # Don't pass the 'fields' arg up to the superclass
        fields = kwargs.pop('fields', None)
        # make sure this is a model class, not a string
        model = kwargs.pop('model', None)
        self.validateRequired = kwargs.pop('required', None)

        if model:
            self.Meta.model = model

        # Instantiate the superclass normally
        super(DynamicLookupModelSerializer, self).__init__(*args, **kwargs)

        if fields:
            # Drop any fields that are not specified in the `fields` argument.
            allowed = set(fields)
            existing = set(self.fields.keys())
            for field_name in existing - allowed:
                self.fields.pop(field_name)

    def to_internal_value(self, data):
        """to internal value.

        :param data:
        :return ret_dict:
        """
        ret_dict = {}
        required = self.validateRequired
        if not required:
            required = ['id']
        for k in required:
            # Perform the data validation.
            val = data.get(k, None)
            if not val:
                raise ValidationError({
                    k: 'This field is required.'
                })
            ret_dict[k] = val

        return ret_dict

    def to_representation(self, instance):
        """to representation.

        :param instance:
        :return ret_dict:
        """
        ret_dict = {}
        if self.fields:
            for k in self.fields:
                ret_dict[k] = getattr(instance, k)

        return ret_dict


class DynamicFieldsModelSerializer(serializers.ModelSerializer):
    """class DynamicFieldsModelSerializer.

    A ModelSerializer that takes an additional `fields` argument that
    controls which fields should be displayed.
    """

    def __init__(self, *args, **kwargs):
        """init.

        :param *args:
        :param **kwargs:
        """
        # Instantiate the superclass normally
        super(DynamicFieldsModelSerializer, self).__init__(*args, **kwargs)

        fields = self.context['request'].query_params.get('fields')
        if fields:
            fields = fields.split(',')
            # Drop any fields that are not specified in the `fields` argument.
            allowed = set(fields)
            existing = set(self.fields.keys())
            for field_name in existing - allowed:
                self.fields.pop(field_name)


class ProjectAdminField(RelatedField):
    """class ProjectAdminField."""

    def to_representation(self, value):
        """to representation.

        :param value:
        :return dict:
        """
        ret_list = []
        for id in value.project_ids.all():
            ret_list.append({'system': id.system.key, 'id': id.identifier})

        return {'title': value.title, 'ids': ret_list, 'id': value.id}



# coding=utf-8
"""

"""
from django.views.decorators.csrf import csrf_exempt
from django.http import JsonResponse


@csrf_exempt
def rapid_connect_auth_view(request):
    ret_dict = dict()
    for key, value in request.POST.items():
        ret_dict[key] = value
    return JsonResponse(ret_dict)



from django.contrib.auth import get_user_model
from hypothesis.extra.django import TestCase
from mixer.backend.django import mixer
from rest_framework.test import APIClient

from crams_contact.serializers.contact_erb_roles import ContactErbRoleSerializer
from crams.models import CramsToken
from crams.utils import crams_utils

User = get_user_model()


class CommonBaseTestCase(TestCase):

    def non_auth_client(self):
        return APIClient()

    def auth_client(self, user=None, token=None):
        # authenticate user and init cramstoken
        client = APIClient()
        if not user:
            user = self.generate_new_user()
        # force user to login
        client.force_authenticate(user=user)
        if not token:
            token = ContactErbRoleSerializer.setup_crams_token_and_roles(user)

        client.credentials(HTTP_AUTHORIZATION='Token ' + token.key)
        return client

    def generate_new_user(self):
        random_str = crams_utils.get_random_string(8)
        username = random_str
        email = username + '@crams.org'
        password = crams_utils.get_random_num_str(10)
        return mixer.blend(User, username=username, password=password, email=email)



# coding=utf-8
"""
Date Utility Functions
"""
import pytz
from calendar import timegm
from merc_common.settings import TIME_ZONE
from django.utils import timezone as d_timezone
from datetime import datetime, date, timedelta, MINYEAR, MAXYEAR
from dateutil.parser import parse
from dateutil.relativedelta import relativedelta


def parse_datetime_from_str(str_datetime_obj):
    if isinstance(str_datetime_obj, datetime):
        return str_datetime_obj
    if isinstance(str_datetime_obj, date):
        return date_to_datetime(str_datetime_obj)
    return convert_naive_to_timezone(parse(str_datetime_obj))


def is_timestamp(obj):
    return isinstance(obj, datetime)


def get_epoch_time(datetime_obj):
    """
        Return Epoch Time
    :param datetime_obj:
    :return:
    """
    return timegm(datetime_obj.timetuple())


def test_get_epoch_time():
    """
        Test get_epoch_time

    """
    assert get_epoch_time(date(1995, 1, 12)) == 789868800


def get_date_str(obj):
    """
        Convert date object to string
    :param obj:
    :return:
    """
    date_obj = obj
    if isinstance(obj, datetime):
        date_obj = obj.date()
    return str(date_obj)


def test_get_date_str():
    """
        Test get_date_str

    """
    assert get_date_str(date(2014, 10, 16)) == '2014-10-16'
    assert get_date_str(get_current_time_for_app_tz()) == get_current_date()


def is_current_year(date_time_obj):
    """
        check if the given date is in current year

    :param date_time_obj:
    :rtype : bool
    """
    return get_current_year() == date_time_obj.year


def test_is_current_year():
    """
        test is_current_year

    """
    now = datetime.now()
    assert is_current_year(now), "tests for current year fails"
    assert not is_current_year(increase_date(
        now.date(), 370)), "next year should be equal to current year"
    assert not is_current_year(reduce_date(now.date(),
                                           370)), \
        "previous year should be equal to current year"


def get_current_year():
    """
        return current year

    :return:
    """
    return get_current_time_for_app_tz().year


def test_get_current_year():
    """
        test get_current_year

    """
    assert datetime.now().year == get_current_year(), 'current year not ' \
                                                      'fetched correctly'


def get_jan01_current_year():
    """
        get First day of Current Year

    :return:
    """
    return date(get_current_year(), 1, 1)


def get_min_date():
    """
    test get_jan01_current_year

    :return:
    """
    return date(MINYEAR, 1, 1)


def get_max_date():
    """
        get system maximum date

    :return:
    """
    return date(MAXYEAR, 12, 31)


def get_current_time_for_app_tz():
    """
        get current time

    :return:
    """
    tz = pytz.timezone(TIME_ZONE)
    return datetime.now(tz)


def get_current_date():
    """
        get current date

    :return:
    """
    return get_current_time_for_app_tz().date()


def date_to_datetime(in_date, choose_end_of_day=False):
    ret_date = None
    if isinstance(in_date, date):
        ret_date = in_date
    elif isinstance(in_date, datetime):
        ret_date = in_date.date()
    elif isinstance(in_date, str):
        ts = parse_datetime_from_str(in_date)
        ret_date = ts.date()

    if choose_end_of_day:
        ret_date = increase_date(ret_date, 1)

    ret_ts = datetime.combine(ret_date, datetime.min.time())
    if choose_end_of_day:
        one_second = timedelta(seconds=1)
        ret_ts = ret_ts - one_second
    return convert_naive_to_timezone(ret_ts)


def get_mssql_date(in_date):
    """
        get date suitable for use by microsoft SQL
    :param in_date:
    :return:
    """
    return in_date.strftime('%Y%m%d')


def get_mssql_timestamp(in_time_stamp):
    """
        get timestamp suitable for use by microsoft SQL
    :param in_time_stamp:
    :return:
    """
    return in_time_stamp.strftime('%Y-%m-%d %H:%M:%S')


def get_day_difference(date_obj0, date_obj1):
    """
        calculate date difference between date_obj0 and date_obj1
    :param date_obj0:
    :param date_obj1:
    :return:
    """
    delta = date_obj0 - date_obj1
    return delta.days


def test_get_day_difference():
    """
        test get_day_difference

    """
    assert get_day_difference(date(2014, 10, 16), date(
        2014, 10, 5)) == 11, "difference in days not computed properly"


def increase_date(date_to_increase, increase_by):
    """
        increase dateObj by given days
    :param date_to_increase:
    :param increase_by:
    :return:
    """
    return transpose_date(date_to_increase, increase_by)


def test_increase_date():
    """
        test increase_date

    """
    assert increase_date(date(2014, 10, 16), 4) == date(
        2014, 10, 20), "increasing by 4 days does not work"


def get_midnight_next_day(timestamp_obj):
    expiry_ts = timestamp_obj + timedelta(days=2)
    return expiry_ts.replace(hour=0, minute=0, second=0, microsecond=0)


def reduce_date(date_to_reduce, reduce_by):
    """
        reduce date obj by given days
    :param date_to_reduce:
    :param reduce_by:
    :return:
    """
    return transpose_date(date_to_reduce, -1 * reduce_by)


def test_reduce_date():
    """
        test reduce_date

    """
    assert reduce_date(date(2014, 10, 16), 3) == date(
        2014, 10, 13), "reducing by 3 days does not work"


def update_by_year(date_to_transpose, year_count):
    return date_to_transpose + relativedelta(years=year_count)


def update_by_month(date_to_transpose, month_count):
    return date_to_transpose + relativedelta(months=month_count)


def transpose_date(date_to_transpose, by_days):
    """
        transpose given date bu given days,
            positive and negative values allowed for by_days parameter
    :param date_to_transpose:
    :param by_days:
    :return:
    """
    return date_to_transpose + timedelta(days=by_days)


def test_transpose_date():
    """
        test transpose_date

    """
    assert transpose_date(date(2014, 10, 16), -6) == date(2014, 10, 10), \
        "transposing by negative 6 days does not work"
    assert transpose_date(date(2014, 10, 16), 6) == date(2014, 10, 22), \
        "transposing by positive 6 days does not work"


def is_given_date_after_current_date(given_date):
    """
        test if given date obj is after current date
    :param given_date:
    :return:
    """
    return get_day_difference(given_date, get_current_date()) > 0


def test_is_current_date_after_given_date():
    """
        test is_current_date_after_current_date

    """
    assert is_given_date_after_current_date(
        increase_date(
            get_current_date(),
            3)), "Current date must be before future date " + str(
        increase_date(
            get_current_date(), 3))
    assert not is_given_date_after_current_date(
        get_current_date()), "Current date must not be before current date"
    assert not is_given_date_after_current_date(
        reduce_date(
            get_current_date(),
            3)), "Current date must not be before past date " + str(
        reduce_date(
            get_current_date(), 3))


def get_minutes_elapsed(end_ts, start_ts):
    return int(get_seconds_elapsed(end_ts, start_ts) / 60)


def get_seconds_elapsed(end_ts, start_ts):
    start_ts_non_naive = convert_naive_to_timezone(start_ts)
    end_ts_non_naive = convert_naive_to_timezone(end_ts)
    elapsed_dt = end_ts_non_naive - start_ts_non_naive
    return elapsed_dt.total_seconds()


def convert_naive_to_timezone(ts, time_zone=TIME_ZONE):
    if d_timezone.is_naive(ts):
        tz = pytz.timezone(time_zone)
        zone_ts = tz.localize(ts)
        return zone_ts
    return ts



# coding=utf-8
"""
Crams Models
"""
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _
from django.core.validators import MinValueValidator
from django.contrib.auth import models as auth_models
from django.db import models
from rest_framework.authtoken.models import Token
from crams.utils import date_utils
from merc_common import settings
from crams.user_manager import UserManager


class CramsToken(Token):
    """
    CramsToken Model
    """
    ks_roles = models.TextField(null=True, blank=True)

    class Meta:
        app_label = 'crams'

    def is_expired(self):
        current_ts = date_utils.get_current_time_for_app_tz()
        elapsed_seconds = date_utils.get_seconds_elapsed(current_ts, self.created)
        return elapsed_seconds >= settings.TOKEN_EXPIRY_TIME_SECONDS


class User(auth_models.AbstractBaseUser, auth_models.PermissionsMixin):
    username = models.CharField(max_length=128, unique=True)
    email = models.EmailField(max_length=75)
    keystone_uuid = models.CharField(max_length=64, blank=True, null=True)
    is_staff = models.BooleanField(_('staff status'), default=False,
                                   help_text=_('Designates whether the user can log into this admin site.'))
    is_active = models.BooleanField(
        _('active'),
        default=True,
        help_text=_('Designates whether this user should be treated as active. '
                    'Unselect this instead of deleting accounts.'))
    first_name = models.CharField(_('first name'),
                                  max_length=30, blank=True, null=True)
    last_name = models.CharField(_('last name'),
                                 max_length=30, blank=True, null=True)
    date_joined = models.DateTimeField(_('date joined'), default=timezone.now)

    objects = UserManager()

    USERNAME_FIELD = 'username'
    REQUIRED_FIELDS = ['email']

    def get_full_name(self):
        # The user is identified by their email address
        return self.email

    def get_short_name(self):
        # The user is identified by their email address
        return self.email

    def __str__(self):  # __unicode__ on Python 2
        return self.email

    def has_perm(self, perm, obj=None):
        """Does the user have a specific permission?"""
        # Simplest possible answer: Yes, always
        return True

    def has_module_perms(self, app_label):
        """Does the user have permissions to view the app `app_label`?"""
        # Simplest possible answer: Yes, always
        return True

    class Meta:
        app_label = 'crams'


class CramsCommon(models.Model):
    """
    CramsCommon Model
    """
    creation_ts = models.DateTimeField(
        auto_now_add=True,
        editable=False
    )
    last_modified_ts = models.DateTimeField(
        auto_now=True,
        editable=False
    )
    created_by = models.ForeignKey(
        User, related_name="%(class)s_created_by", blank=True, null=True, on_delete=models.DO_NOTHING)
    updated_by = models.ForeignKey(
        User, related_name="%(class)s_updated_by", blank=True, null=True, on_delete=models.DO_NOTHING)

    class Meta:
        abstract = True
        app_label = 'crams'


class ArchivableModel(models.Model):
    archive_ts = models.DateTimeField(blank=True, null=True)

    class Meta:
        abstract = True
        app_label = 'crams'

    def archive_instance(self):
        self.archive_ts = date_utils.get_current_time_for_app_tz()
        self.save()

    @classmethod
    def get_current_for_qs(cls, archivable_qs):
        if isinstance(archivable_qs.model, type(ArchivableModel)):
            return archivable_qs.filter(archive_ts__isnull=True)


class UserEvents(CramsCommon):
    """
    UserEvents Model
    """
    event_message = models.TextField()

    related_user = models.ForeignKey(User, blank=True, null=True, on_delete=models.DO_NOTHING)

    class Meta:
        app_label = 'crams'


class Provider(CramsCommon):
    """
    Provider Model
    Each provider object in crams represents an infrastructure provisioning team.
    The membership of this provisioning team is left undefined in this module.
    Instead it is expected that the provider object is associated with a set of eResearch Body systems
    """
    name = models.CharField(max_length=200, unique=False)
    start_date = models.DateField(auto_now_add=True, editable=False)
    active = models.BooleanField(default=True)
    description = models.TextField(null=True, blank=True)

    @property
    def provider_role(self):
        return self.name.strip().lower() + '_provider'

    def get_status_str(self):
        """

        get status as string
        :return:
        """
        if not self.active:
            return 'inActive'
        return 'current'

    class Meta:
        app_label = 'crams'

    def __str__(self):
        return '{}'.format(self.name)


class ProvisionDetails(CramsCommon):
    """
    ProvisionDetails Model
    The status flow is
    - From Nothing to Sent
    - From Sent to Provisioned or Failed
    - From Provisioned to Updated
    - From Failed to Updated
    - From Updated to Update Sent
    - From Update Sent to Provisioned or Failed
    """
    UNKNOWN = '?'
    SENT = 'S'
    PROVISIONED = 'P'
    POST_PROVISION_UPDATE = 'U'
    POST_PROVISION_UPDATE_SENT = 'X'
    FAILED = 'F'
    RESEND_LATER = 'L'
    SET_OF_SENT = frozenset([SENT, POST_PROVISION_UPDATE_SENT])
    READY_TO_SEND_SET = frozenset([RESEND_LATER, POST_PROVISION_UPDATE])
    STATUS_CHOICES = ((SENT, 'Sent'), (PROVISIONED, 'Provisioned'),
                      (FAILED, 'Failed'), (RESEND_LATER, 'Resend'),
                      (POST_PROVISION_UPDATE, 'Updated'),
                      (POST_PROVISION_UPDATE_SENT, 'Update Sent'),
                      (UNKNOWN, '???'))
    status = models.CharField(
        max_length=1, choices=STATUS_CHOICES, default=SENT)
    message = models.TextField(blank=True, null=True)
    parent_provision_details = models.ForeignKey(
        'ProvisionDetails', blank=True, null=True, related_name='history', on_delete=models.DO_NOTHING)

    class Meta:
        app_label = 'crams'

    def __str__(self):
        return '{}. {} : {}'.format(self.id, self.status, self.message)


class ProvisionableItem(models.Model):
    """
    ProvisionableItem Model
    """
    provision_details = models.OneToOneField(
        ProvisionDetails, blank=True, null=True, related_name='%(class)s', on_delete=models.DO_NOTHING)

    class Meta:
        abstract = True
        app_label = 'crams'

    def get_provider(self):
        """
        get provider
        :raise NotImplementedError:
        """
        raise NotImplementedError(
            'Get Provider not implemented for abstract Product Request model')


class EResearchBody(models.Model):
    name = models.CharField(max_length=200, db_index=True)

    email = models.EmailField(null=True, blank=True)

    class Meta:
        app_label = 'crams'
        # indexes = [
        #     models.Index(fields=['name'], name='erb_name_idx'),
        # ]

    def __str__(self):
        return self.name

    @property
    def admin_role(self):
        return self.name.strip().lower() + '_erb_admin'

    @property
    def tenancy_manager_role(self):
        return self.name.strip().lower() + '_erb_tenancy_manager'


class EResearchBodySystem(models.Model):
    e_research_body = models.ForeignKey(EResearchBody, related_name='systems', on_delete=models.DO_NOTHING)

    name = models.CharField(max_length=200, db_index=True)

    email = models.EmailField(null=True, blank=True)

    class Meta:
        app_label = 'crams'
        unique_together = ('name', 'e_research_body')
        index_together = ['name', 'e_research_body']
        # indexes = [
        #     models.Index(fields=['name', 'e_research_body']),
        #     models.Index(fields=['name'], name='erb_system_name_idx'),
        # ]

    def __str__(self):
        return self.name

    @property
    def admin_role(self):
        return self.name.strip().lower() + '_erbs_admin'

    @property
    def approver_role(self):
        return self.name.strip().lower() + '_erbs_approver'


class EResearchBodyDelegate(models.Model):
    e_research_body = models.ForeignKey(
        EResearchBody, related_name='delegates', on_delete=models.DO_NOTHING)

    name = models.CharField(max_length=200, db_index=True)

    email = models.EmailField(null=True, blank=True)

    class Meta:
        app_label = 'crams'
        unique_together = ('name', 'e_research_body')
        index_together = ['name', 'e_research_body']
        # indexes = [
        #     models.Index(fields=['name', 'e_research_body']),
        #     models.Index(fields=['name'], name='erb_delegate_name_idx'),
        # ]

    def __str__(self):
        return self.name

    @property
    def approver_role(self):
        return self.name.strip().lower() + '_delegate_approver'


class EResearchBodyIDKey(models.Model):
    """
    EResearchBodyIDKey Model
    """
    ID_KEY = 'I'
    LABEL = 'L'
    METADATA = 'M'
    TYPE_CHOICES = ((ID_KEY, 'Id_Key'), (LABEL, 'Label'),
                    (METADATA, 'Metadata'))
    type = models.CharField(
        max_length=1, choices=TYPE_CHOICES, default=ID_KEY)

    key = models.CharField(max_length=100)

    e_research_body = models.ForeignKey(EResearchBody, blank=True, null=True, on_delete=models.DO_NOTHING)

    description = models.TextField(blank=True, null=True)

    class Meta:
        app_label = 'crams'
        unique_together = ('key', 'e_research_body')

    def __str__(self):
        return '{} / {}'.format(self.key, self.e_research_body)


class Question(models.Model):
    """
    Question Model
    """
    key = models.CharField(
        max_length=50, unique=True
    )

    question_type = models.CharField(
        max_length=200
    )

    question = models.CharField(
        max_length=200
    )

    class Meta:
        app_label = 'crams'

    def __str__(self):
        return '{} {} {}'.format(self.key, self.question_type, self.question)


class FundingBody(models.Model):
    """
    FundingBody Model
    """
    name = models.CharField(
        max_length=200
    )

    email = models.EmailField(null=True, blank=True)

    class Meta:
        app_label = 'crams'

    def __str__(self):
        return '{} - {}'.format(self.name, self.email)


class FundingScheme(models.Model):
    """
    FundingScheme Model
    """
    funding_scheme = models.CharField(
        max_length=200
    )

    funding_body = models.ForeignKey(
        FundingBody, related_name='funding_schemes', on_delete=models.DO_NOTHING)

    class Meta:
        app_label = 'crams'

    def __str__(self):
        return '{} {}'.format(self.funding_body.name, self.funding_scheme)


class Duration(models.Model):
    """
    Duration Model
    """
    duration = models.IntegerField(
        default=1,
        validators=[MinValueValidator(1)]
    )
    duration_label = models.CharField(
        max_length=50
    )

    class Meta:
        app_label = 'crams'

    def __str__(self):
        return '{} {}'.format(self.duration, self.duration_label)


class FORCode(models.Model):
    """
    FORCode Model
    """
    code = models.CharField(
        max_length=50
    )

    description = models.CharField(
        max_length=200
    )

    class Meta:
        app_label = 'crams'

    def __str__(self):
        return '{} {}'.format(self.code, self.description)


class SEOCode(models.Model):
    code = models.CharField(max_length=200)
    description = models.CharField(max_length=200)

    class Meta:
        app_label = 'crams'

    def __str__(self):  # __unicode__ on Python 2
        return self.__unicode__()

    def __unicode__(self):
        return str(self.code)


class CramsSequence(models.Model):
    code = models.CharField(max_length=255, primary_key=True)
    last_sequence_num = models.IntegerField(default=0)

    class Meta:
        app_label = 'crams'

    def __str__(self):  # __unicode__ on Python 2
        return self.__unicode__()

    def __unicode__(self):
        return str(self.code + ' : ' + str(self.last_sequence_num))

    @classmethod
    def get_code_increment_sequence(cls, code):
        seq, created = CramsSequence.objects.get_or_create(code=code)
        seq.last_sequence_num += 1
        seq.save()
        return seq


class SupportEmailContact(models.Model):
    description = models.CharField(max_length=255, blank=True, null=True)
    email = models.EmailField(db_index=True)
    e_research_body = models.ForeignKey(EResearchBody, on_delete=models.DO_NOTHING)
    e_research_system = models.ForeignKey(
        EResearchBodySystem, null=True, blank=True, on_delete=models.DO_NOTHING)

    class Meta:
        app_label = 'crams'

    def __str__(self):
        return '{} {}'.format(self.description, self.email)


class MessageOfTheDay(CramsCommon):
    message = models.CharField(max_length=1500, blank=True, null=True)
    e_research_body = models.ForeignKey(EResearchBody, null=True, blank=True, on_delete=models.DO_NOTHING)

    class Meta:
        app_label = 'crams'

    def __str__(self):
        return '{}'.format(self.message)


class CostUnitType(models.Model):
    unit_name = models.CharField(max_length=19)

    class Meta:
        app_label = 'crams'

    def __str__(self):
        return self.unit_name


class ProductCommon(models.Model):
    """
    ProductCommon Model
    """
    name = models.CharField(max_length=200)

    funding_body = models.ForeignKey(FundingBody, related_name='%(class)s', on_delete=models.DO_NOTHING)

    e_research_system = models.ForeignKey(
        EResearchBodySystem, related_name='%(class)s',
        blank=True, null=True, default=None, on_delete=models.DO_NOTHING)

    provider = models.ForeignKey(Provider, related_name='%(class)s', on_delete=models.DO_NOTHING)

    unit_cost = models.FloatField(default=0.0)

    operational_cost = models.FloatField(default=0.0)

    cost_unit_type = models.ForeignKey(CostUnitType, related_name='%(class)s',
                                       blank=True, null=True, on_delete=models.DO_NOTHING)

    capacity = models.FloatField(default=0.0)

    class Meta:
        abstract = True
        unique_together = ('name', 'e_research_system')
        index_together = ['name', 'e_research_system']
        app_label = 'crams'

    def __str__(self):
        return '{} {}'.format(self.id, self.name)


class Zone(models.Model):
    """
    Zone Model
    """
    name = models.CharField(max_length=64)
    description = models.TextField()

    class Meta:
        app_label = 'crams'

    def __str__(self):
        return '{}'.format(self.name)


class AbstractAllocation(ProvisionableItem):
    """
    AbstractAllocation Model
    """
    current_quota = models.FloatField(default=0.0)

    requested_quota_change = models.FloatField(default=0.0)

    approved_quota_change = models.FloatField(default=0.0)

    class Meta:
        app_label = 'crams_storage'
        abstract = True

    @property
    def requested_quota_total(self):
        return self.current_quota + self.requested_quota_change

    @property
    def approved_quota_total(self):
        return self.current_quota + self.approved_quota_change


class ProjectMemberStatus(models.Model):
    """
    ProjectMemberStatus Model
    """
    code = models.CharField(
        max_length=50
    )

    status = models.CharField(
        max_length=100
    )

    class Meta:
        app_label = 'crams'

    def __str__(self):
        return '{} {} {}'.format(self.id, self.code, self.status)


class SoftwareLicenseStatus(models.Model):
    code = models.CharField(max_length=50)
    status = models.CharField(max_length=100)

    class Meta:
        app_label = 'crams'

    def __str__(self):
        return '{} {} {}'.format(self.id, self.code, self.status)


class AbstractNotificationTemplate(models.Model):
    funding_body = models.ForeignKey(FundingBody, blank=True, null=True, default=None,
                                     related_name='notification_templates', on_delete=models.DO_NOTHING)
    system_key = models.ForeignKey(EResearchBodyIDKey,
                                   blank=True, null=True, default=None, on_delete=models.DO_NOTHING)
    template_file_path = models.CharField(max_length=99)
    alert_funding_body = models.BooleanField(default=False)
    e_research_system = models.ForeignKey(
        EResearchBodySystem, blank=True, null=True, default=None,
        related_name='notification_templates', on_delete=models.DO_NOTHING)
    e_research_body = models.ForeignKey(
        EResearchBody, blank=True, null=True, default=None, on_delete=models.DO_NOTHING)

    class Meta:
        abstract = True

    def __str__(self):
        return '{} {}'.format(self.template_file_path, self.e_research_system)



# coding=utf-8
"""

"""
import abc

from rest_framework import serializers
from rest_framework.exceptions import ParseError

from crams.utils.role import AbstractCramsRoleUtils as RoleUtils

from crams import models
from crams.serializers.model_serializers import ReadOnlyModelSerializer


class ReadOnlyProvisionDetailSerializer(ReadOnlyModelSerializer):
    """
    ==> from d3_prod crams.common.serializers.provision_detail_serializers
    """
    class Meta(object):
        model = models.ProvisionDetails
        fields = ('status', 'message')


class AbstractProvisionDetailsSerializer(serializers.ModelSerializer):
    """
    ==> from d3_prod crams.api.v1.serializers.utilitySerializers
    """
    """class ProvisionDetailsSerializer."""

    SHOW_FAIL_MESSAGE = 'show_fail_message'

    class Meta(object):
        model = models.ProvisionDetails
        abstract = True
        fields = ('id', 'status', 'message')

    @classmethod
    def hide_error_msg_context(cls):
        return {cls.SHOW_FAIL_MESSAGE: False}

    @classmethod
    def show_error_msg_context(cls):
        return {cls.SHOW_FAIL_MESSAGE: True}

    @classmethod
    def build_context_obj(cls, user_obj, erb_system_obj=None):
        if RoleUtils.has_e_system_role(user_obj, erb_system_obj):
            return cls.show_error_msg_context()
        return cls.hide_error_msg_context()

    @classmethod
    def sanitize_provision_details_for_user(cls, provision_details_dict):
        if provision_details_dict.get('status') == models.ProvisionDetails.FAILED:
            provision_details_dict['status'] = models.ProvisionDetails.SENT
            provision_details_dict['message'] = None

    def to_representation(self, instance):
        data = super().to_representation(instance)
        if self.context:
            if not self.context.get(self.SHOW_FAIL_MESSAGE, False):
                self.sanitize_provision_details_for_user(data)
        return data

    @abc.abstractmethod
    def _get_new_provision_status(self, existing_status):
        pass

    def create(self, validated_data):
        p_status = validated_data.get('status')

        validated_data['status'] = self._get_new_provision_status(p_status)

        return models.ProvisionDetails.objects.create(**validated_data)

    def update(self, instance, validated_data):
        """

        :param instance:
        :param validated_data:
        :return:
        """
        raise ParseError('models.ProvisionDetails: Update not allowed, clone new ')



# coding=utf-8
"""

"""


class CramsAspect:
    def __init__(self, serializer_class):
        self.serializer_class = serializer_class
        self.pre_method_dict = dict()
        self.post_method_dict = dict()

    def add_pre_fn_for_method_name(self, method_name, pre_fn):
        if method_name not in self.pre_method_dict:
            self.pre_method_dict[method_name] = set()
        upd_pre_fn_set = self.pre_method_dict[method_name]
        upd_pre_fn_set.add(pre_fn)

    def add_post_fn_for_method_name(self, method_name, post_fn):
        if method_name not in self.post_method_dict:
            self.post_method_dict[method_name] = set()
        upd_post_fn_set = self.post_method_dict[method_name]
        upd_post_fn_set.add(post_fn)


CRAMS_ASPECT_DICT = dict()



from rest_framework import viewsets, exceptions, mixins

from crams import models
from crams.utils import rest_utils


class LookupViewset(viewsets.ReadOnlyModelViewSet):
    def get_current_user(self, http_request=None):
        if http_request:
            return http_request.user
        user, _ = rest_utils.get_current_user_from_context(self)
        return user

    def fetch_e_research_body_param_obj(self, erb_name=None):
        if not erb_name:
            erb_name = self.request.query_params.get('e_research_body', None)
        if erb_name:
            try:
                return models.EResearchBody.objects.get(name__iexact=erb_name)
            except models.EResearchBody.DoesNotExist:
                msg = 'EResearchBody with name {} does not exist'
                raise exceptions.ValidationError(msg.format(erb_name))
        return None

    def fetch_system_key_param_obj(self, erb_obj=None, key=None, type=None):
        erb_obj = erb_obj or self.fetch_e_research_body_param_obj()
        if erb_obj:
            key = key or self.request.query_params.get('key', None)
            qs = models.EResearchBodyIDKey.objects.filter(
                        e_research_body=erb_obj, key__iexact=key)
            if type:
                qs = qs.filter(type=type)

            if not qs.exists():
                prefix = 'System ID'
                if type == models.EResearchBodyIDKey.METADATA:
                    prefix = 'Meta label'
                elif type == models.EResearchBodyIDKey.LABEL:
                    prefix = 'Label'
                msg = '{} with key {} does not exist for {}'
                raise exceptions.ValidationError(
                    msg.format(prefix, key, erb_obj))

            return qs.first()
        return None


class CreateListViewset(LookupViewset, mixins.CreateModelMixin):
    pass


class UpdateListViewset(CreateListViewset, mixins.UpdateModelMixin):
    pass



# Generated by Django 3.2.4 on 2021-07-21 01:19

from django.db import migrations, models

def load_software_status_sql():
    import os

    file_name = 'software_status.sql'
    sql_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), '../sqls'))
    sql_statements = open(os.path.join(sql_dir, file_name), 'r').read()

    return sql_statements

class Migration(migrations.Migration):

    dependencies = [
        ('crams', '0010_projectmemberstatus_lookups'),
    ]

    operations = [
        migrations.CreateModel(
            name='SoftwareLicenseStatus',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('code', models.CharField(max_length=50)),
                ('status', models.CharField(max_length=100)),
            ],
        ),
        migrations.RunSQL(load_software_status_sql()),
    ]



# coding=utf-8
"""
    DB Constants
"""

JSON_APPROVER_STR = 'approver'

# SystemIds
SYSTEM_NECTAR_UUID = 'NeCTAR_UUID'
SYSTEM_NECTAR_CREATED_BY = 'NeCTAR_Created_By'
SYSTEM_NECTAR_DB_ID = 'NeCTAR_DB_Id'
SYSTEM_NECTAR = 'NeCTAR'
SYSTEM_VICNODE = 'VicNode'
SYSTEM_RACMON_SONAS = 'SONAS'
SYSTEM_HPC = 'HPC'

# eResearch Body
eRESEARCH_BODY_RACMON = 'RaCMon'
eRESEARCH_BODY_HPC = 'HPC'

# eResearch Systems
eSYSTEM_NECTAR = 'NeCTAR'
eSYSTEM_RACMON_MONASH = 'RaCMon-Monash'
eSYSTEM_VICNODE = 'VicNode'
eSYSTEM_MASSIVE = 'MASSIVE'
eSYSTEM_CVL = 'CVL'
eSYSTEM_MONARCH = 'MonARCH'

# Allocation Home
ALLOCATION_HOME_MONASH = 'Monash'

# ERB IdKeys
ERB_ID_NECTAR_TENANT_ID = 'NECTAR_TENANT_ID'

# Contact Role
APPLICANT = 'Applicant'
CHIEF_INVESTIGATOR = 'Chief Investigator'
INVITEE = 'Invitee'
DATA_CUSTODIAN = 'Data Custodian'
TECHNICAL_CONTACT = 'Technical Contact'

# ERB IDs
HPC_POSIX_USERID = 'HPC_POSIX_USERID'
HPC_POSIX_GROUP_ID = 'HPC_POSIX_GROUP_ID'
HPC_SOFTWARE_GROUP_ID = 'HPC_SOFTWARE_GROUP_ID'

# Request Status Codes
REQUEST_STATUS_NEW = 'N'
REQUEST_STATUS_SUBMITTED = 'E'
REQUEST_STATUS_UPDATE_OR_EXTEND = 'X'
REQUEST_STATUS_APPROVED = 'A'
REQUEST_STATUS_DECLINED = 'R'
REQUEST_STATUS_PARTIAL_PROVISIONED = '_PP'
REQUEST_STATUS_PROVISIONED = 'P'
REQUEST_STATUS_UPDATE_OR_EXTEND_DECLINED = 'J'
REQUEST_STATUS_LEGACY_SUB = 'L'
REQUEST_STATUS_LEGACY_APPROVED = 'M'
REQUEST_STATUS_LEGACY_DECLINED = 'O'
REQUEST_STATUS_APPLICATION_UPDATED = 'U'

REQUEST_STATUS_EXTEND_DRAFT = 'D'
DRAFT_PERMISSIBLE_STATES = [REQUEST_STATUS_NEW,
                            REQUEST_STATUS_PROVISIONED,
                            REQUEST_STATUS_EXTEND_DRAFT]
DRAFT_STATES = [REQUEST_STATUS_NEW, REQUEST_STATUS_EXTEND_DRAFT]
EDIT_STATUS = DRAFT_STATES + [REQUEST_STATUS_SUBMITTED,
                              REQUEST_STATUS_DECLINED,
                              REQUEST_STATUS_UPDATE_OR_EXTEND_DECLINED,
                              REQUEST_STATUS_UPDATE_OR_EXTEND]

LEGACY_STATES = [REQUEST_STATUS_LEGACY_DECLINED,
                 REQUEST_STATUS_LEGACY_APPROVED, REQUEST_STATUS_LEGACY_SUB]
NON_ADMIN_STATES = [REQUEST_STATUS_NEW, REQUEST_STATUS_SUBMITTED,
                    REQUEST_STATUS_UPDATE_OR_EXTEND, REQUEST_STATUS_LEGACY_SUB]

APPROVAL_STATES = [REQUEST_STATUS_APPROVED, REQUEST_STATUS_LEGACY_APPROVED]

NEW_REQUEST_STATUS = [REQUEST_STATUS_NEW, REQUEST_STATUS_SUBMITTED, REQUEST_STATUS_DECLINED]
DECLINED_STATES = [
    REQUEST_STATUS_DECLINED,
    REQUEST_STATUS_UPDATE_OR_EXTEND_DECLINED,
    REQUEST_STATUS_LEGACY_DECLINED]
ADMIN_STATES = APPROVAL_STATES + DECLINED_STATES + [REQUEST_STATUS_PROVISIONED]
QUOTA_CHANGE_STATES = ADMIN_STATES  # approved + declined + provisioned states

# Notification Contact ROLEs
E_RESEARCH_BODY_ADMIN = 'E_RESEARCH_BODY Admin'
E_RESEARCH_BODY_SYSTEM_ADMIN = 'E_RESEARCH_BODY_SYSTEM Admin'
E_RESEARCH_SYSTEM_DELEGATE = 'E_RESEARCH_SYSTEM_DELEGATE'
E_RESEARCH_BODY_SYSTEM_APPROVER = 'E_RESEARCH_BODY_SYSTEM_Approver'
E_RESEARCH_BODY_SYSTEM_PROVISIONER = 'E_RESEARCH_BODY_SYSTEM_Provisioner'

NOTIFY_CONTACT = 'NOTIFY_CONTACT'
NOTIFY_PROJECT_LEADER ='NOTIFY_PROJECT_LEADER'
NOTIFY_NON_READONLY_PROJECT_CONTACT = 'NOTIFY_NON_READONLY_PROJECT_CONTACT'
NOTIFY_READ_ONLY_PROJECT_CONTACT = 'NOTIFY_READ_ONLY_PROJECT_CONTACT'

PRODUCT_NAME_MAPPING = []




from celery import task

from crams.mail import mail_sender


@task(name="send_notification")
def send_email_notification(
        sender,
        subject,
        mail_content,
        recipient_list,
        template_name=None,
        cc_list=None,
        bcc_list=None,
        reply_to=None,
        fail_silently=False):
    mail_sender.send_email(sender, subject, mail_content, recipient_list, template_name, cc_list, bcc_list, reply_to,
                           fail_silently)



# coding=utf-8
"""

"""
from rest_framework.response import Response
from rest_framework import viewsets

from crams.permissions import IsCramsAuthenticated
from crams.models import EResearchBodySystem
from crams.utils.role import AbstractCramsRoleUtils


class PrevUserFundingBodyRoleList(viewsets.ViewSet):
    """
     current user approver role list
    """
    permission_classes = [IsCramsAuthenticated]

    # noinspection PyMethodMayBeStatic
    def list(self, request):
        """
            List
        :param request:
        :return:
        """
        # ret_list = self.fetch_user_funding_body_roles(request.user)
        ret_list = self.fetch_user_system_roles(request.user)
        return Response(ret_list)

    @classmethod
    def fetch_user_system_roles(cls, user_obj):
        ret_list = []
        system_list = AbstractCramsRoleUtils.get_authorised_e_research_system_list(user_obj)
        if len(system_list) > 0:
            for es in EResearchBodySystem.objects.filter(name__in=system_list):
                current_dict = {'id': es.id, 'name': es.name}
                ret_list.append(current_dict)
                current_dict[AbstractCramsRoleUtils.APPROVER_ROLE_KEY] = True
        return ret_list



# coding=utf-8
"""Utils."""

import inspect
import itertools
import random
import string
import io


def get_random_num_str(num):
    return ''.join(
        random.SystemRandom().choice(string.digits) for _ in range(num))


def get_random_string(num):
    # http://stackoverflow.com/questions/2257441/random-string-generation-with
    # -upper-case-letters-and-digits-in-python/23728630#23728630
    # return ''.join(random.choice(string.ascii_uppercase + string.digits) for
    # _ in range(num))
    """
        get_random_string
    :param num:
    :return:
    """
    return ''.join(
        random.SystemRandom().choice(
            string.ascii_uppercase +
            string.digits) for _ in range(num))


# Modified from
# http://stackoverflow.com/questions/18826571/python-powerset-of-a-given-set-with-generators
def power_set_generator(input_set):
    """
        power_set_generator
    :param input_set:
    """
    for subset in itertools.chain.from_iterable(
        itertools.combinations(
            input_set, r) for r in range(
            len(input_set) + 1)):
        yield set(subset)


def debug_function_name(level=1, print_upto=False):
    """
    Prints name of the current calling function, useful in debug,
    :return:
    """
    # current function ==> print(inspect.stack()[0][3])
    if print_upto:
        count = 0
        while count < level:
            print(inspect.stack()[count][3])
            count += 1
    print(inspect.stack()[level][3])


def print_to_string(*args, **kwargs):
    output = io.StringIO()
    print(*args, file=output, **kwargs)
    contents = output.getvalue()
    output.close()
    return contents



# Generated by Django 3.1.1 on 2021-03-08 06:43

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('crams', '0006_cramstoken'),
    ]

    operations = [
        migrations.CreateModel(
            name='CostUnitType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('unit_name', models.CharField(max_length=19)),
            ],
        ),
        migrations.CreateModel(
            name='Zone',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=64)),
                ('description', models.TextField()),
            ],
        ),
    ]



# coding=utf-8
"""
API Constants
"""
REQUEST_PARAM_DRAFT = 'draft'
CLONE = 'clone'

# Serializer Override
OVERRIDE_READONLY_DATA = 'OVERRIDE_READONLY'
DO_NOT_OVERRIDE_PROVISION_DETAILS = 'DO_NOT_OVERRIDE_PROVISION_DETAILS'



# Generated by Django 3.1.1 on 2020-09-25 02:20

from django.conf import settings
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('auth', '0012_alter_user_first_name_max_length'),
    ]

    operations = [
        migrations.CreateModel(
            name='User',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('password', models.CharField(max_length=128, verbose_name='password')),
                ('last_login', models.DateTimeField(blank=True, null=True, verbose_name='last login')),
                ('is_superuser', models.BooleanField(default=False, help_text='Designates that this user has all permissions without explicitly assigning them.', verbose_name='superuser status')),
                ('username', models.CharField(max_length=128, unique=True)),
                ('email', models.EmailField(max_length=75)),
                ('keystone_uuid', models.CharField(blank=True, max_length=64, null=True)),
                ('is_staff', models.BooleanField(default=False, help_text='Designates whether the user can log into this admin site.', verbose_name='staff status')),
                ('is_active', models.BooleanField(default=True, help_text='Designates whether this user should be treated as active. Unselect this instead of deleting accounts.', verbose_name='active')),
                ('first_name', models.CharField(blank=True, max_length=30, null=True, verbose_name='first name')),
                ('last_name', models.CharField(blank=True, max_length=30, null=True, verbose_name='last name')),
                ('date_joined', models.DateTimeField(default=django.utils.timezone.now, verbose_name='date joined')),
                ('groups', models.ManyToManyField(blank=True, help_text='The groups this user belongs to. A user will get all permissions granted to each of their groups.', related_name='user_set', related_query_name='user', to='auth.Group', verbose_name='groups')),
                ('user_permissions', models.ManyToManyField(blank=True, help_text='Specific permissions for this user.', related_name='user_set', related_query_name='user', to='auth.Permission', verbose_name='user permissions')),
            ],
        ),
        migrations.CreateModel(
            name='UserEvents',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('creation_ts', models.DateTimeField(auto_now_add=True)),
                ('last_modified_ts', models.DateTimeField(auto_now=True)),
                ('event_message', models.TextField()),
                ('created_by', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='userevents_created_by', to=settings.AUTH_USER_MODEL)),
                ('related_user', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.DO_NOTHING, to=settings.AUTH_USER_MODEL)),
                ('updated_by', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='userevents_updated_by', to=settings.AUTH_USER_MODEL)),
            ],
        ),
    ]



# coding=utf-8
"""Utils."""

from crams.constants import api


class CramsAPIActionState:
    """class CramsAPIActionState."""
    def __init__(self, parent_obj):
        self.user_erb_roles = list()
        self.error_message = None
        if not parent_obj:
            self.error_message = 'CramsActionState: invalid object'
            return

        self.existing_instance = parent_obj.instance
        self.is_update_action = self.existing_instance is not None
        self.is_create_action = not self.is_update_action
        self.is_partial_action = parent_obj.partial
        self.is_clone_action = False
        self.user_erb_roles_dict = dict()

        if not parent_obj.context or 'request' not in parent_obj.context:
            self.error_message = '"context" object not found, required to ' \
                                 + 'identify current user.'
            raise Exception(self.error_message)

        self.rest_request = parent_obj.context.get('request', None)

        self.is_draft_status_requested = False
        if self.rest_request and api.REQUEST_PARAM_DRAFT \
                in self.rest_request.query_params:
            self.is_draft_status_requested = True

        self.override_data = parent_obj.context.get(api.OVERRIDE_READONLY_DATA,
                                                    None)
        if self.override_data:
            if api.CLONE in self.override_data:
                if not self.existing_instance:
                    self.error_message = 'existing_instance required for \
                                         Clone Action'
                    return
                else:
                    self.is_clone_action = True
                    self.is_update_action = False

        # noinspection PyPep8
        self.update_from_existing = self.is_partial_action or \
            self.is_clone_action

    def set_user_erb_roles(self, key, erb_role_list):
        self.user_erb_roles[key] = erb_role_list



# Generated by Django 3.1.1 on 2020-11-20 01:02

import os

from django.db import migrations


def load_for_code_data_from_sql():
    file_name = 'for_code.sql'
    sqls_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), '../sqls'))
    sql_statements = open(os.path.join(sqls_dir, file_name), 'r').read()
    return sql_statements


def load_seo_code_data_from_sql():
    file_name = 'seo_code.sql'
    sqls_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), '../sqls'))
    sql_statements = open(os.path.join(sqls_dir, file_name), 'r').read()

    return sql_statements


def load_project_duration_data_from_sql():
    file_name = 'project_duration.sql'
    sqls_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), '../sqls'))
    sql_statements = open(os.path.join(sqls_dir, file_name), 'r').read()

    return sql_statements


class Migration(migrations.Migration):
    dependencies = [
        ('crams', '0007_costunittype_zone'),
    ]

    operations = [
        migrations.RunSQL(load_project_duration_data_from_sql()),
        migrations.RunSQL(load_for_code_data_from_sql()),
        migrations.RunSQL(load_seo_code_data_from_sql()),
    ]



# coding=utf-8
"""
Question Related Serializers
"""
import copy

from rest_framework import serializers
from rest_framework.exceptions import ValidationError
from crams import models
from crams.serializers import lookup_serializers
from crams.serializers import model_serializers
from crams.utils import validaton_utils


class AbstractQResponseSerializer(model_serializers.CreateOnlyModelSerializer):
    """class AbstractQResponseSerializer."""

    question = serializers.SlugRelatedField(
        slug_field='key', queryset=models.Question.objects.all())

    question_response = serializers.CharField()

    class Meta(object):
        model = None
        fields = ('question_response', 'question')


class AbstractQuestionResponseSerializer(model_serializers.ReadOnlyModelSerializer):
    """class AbstractQuestionResponseSerializer."""
    id = serializers.IntegerField(required=False, read_only=True)

    question = lookup_serializers.QuestionSerializer(required=True)

    question_response = serializers.CharField(required=False, allow_blank=True, allow_null=True)

    class Meta(object):
        model = None
        fields = ('id', 'question_response', 'question')

    def validate(self, attrs):
        attrs = super().validate(attrs)

        question_obj = attrs.get('question')
        if not isinstance(question_obj, models.Question):
            raise ValidationError('Question cannot be determined for: {}'.format(question_obj))

        return attrs

    def save_model_question_response(self):
        """
        """
        question_response = self.validated_data.pop('question_response')
        obj, created = self.Meta.model.objects.get_or_create(**self.validated_data)
        if not obj.question_response == question_response:
            obj.question_response = question_response
            obj.save()
        return obj, created

    @classmethod
    def save_parent_question_response_return_changes(
            cls, question_response_dict_list, fill_parent_fn, old_parent_responses, parent_attr_name):
        """

        :param question_response_dict_list:
        :param fill_parent_fn:
        :param old_parent_responses:
        :param parent_attr_name: one of 'project', 'request', 'compute_request', 'storage_request'
            - required for managing delete/archive of question responses
        :return: question_response_obj_dict, modified_questions_set, new_questions_set, removed_questions_set, err_list
                 - where is question_response_obj_dict is obj_dict[question] = obj
                 - where each questions set is a set of tuple (question, new_response, existing_response)
        """
        ret_obj = validaton_utils.ChangedMetadata()
        existing_pqr_dict = dict()
        if old_parent_responses:
            for pqr in old_parent_responses:
                existing_pqr_dict[pqr.question] = (pqr.question_response, getattr(pqr, parent_attr_name))

        err_list = list()
        sz_list = list()
        if question_response_dict_list:
            for question_response_dict in question_response_dict_list:
                data = copy.copy(question_response_dict)

                sz = cls(data=data)
                sz.is_valid(raise_exception=False)
                if sz.errors:
                    err_list.append(sz.errors)
                else:
                    sz_list.append(sz)
        if err_list:
            ret_obj.err_list = err_list
            return ret_obj

        new_pqr_dict = dict()
        for sz in sz_list:
            question = sz.validated_data.get('question')
            question_response = sz.validated_data.get('question_response')
            new_pqr_dict[question] = (question_response, sz)

        new_questions_set = set()
        modified_questions_set = set()
        removed_questions_set = set()
        all_questions = set(list(existing_pqr_dict.keys()) + list(new_pqr_dict.keys()))
        obj_dict = dict()
        for question in all_questions:
            existing_response, existing_parent = None, None
            existing_tuple = existing_pqr_dict.get(question)
            if existing_tuple:
                existing_response, existing_parent = existing_tuple

            new_response, sz = None, None
            new_tuple = new_pqr_dict.get(question)
            if new_tuple:
                new_response, sz = new_pqr_dict.get(question)
                new_parent = fill_parent_fn(sz.validated_data)
            else:
                new_parent = fill_parent_fn(None)

            if existing_response == new_response:
                if not existing_parent == new_parent:
                    obj, created = sz.save_model_question_response()
                    obj_dict[question] = obj
                continue
            if not existing_response:
                new_questions_set.add((question, new_response, existing_response))
                obj, created = sz.save_model_question_response()
                obj_dict[question] = obj
            elif not new_response:
                removed_questions_set.add((question, new_response, existing_response))
                if existing_parent == new_parent:  # archive only if the is update is to existing parent
                    if isinstance(existing_response, models.ArchivableModel):
                        existing_response.archive_instance()
            else:
                modified_questions_set.add((question, new_response, existing_response))
                obj, created = sz.save_model_question_response()
                obj_dict[question] = obj

        ret_obj.obj_dict = obj_dict
        ret_obj.modified_related_field_set = modified_questions_set
        ret_obj.removed_related_field_set = removed_questions_set
        ret_obj.new_related_field_set = new_questions_set
        return ret_obj


class BaseQuestionResponseSerializer(AbstractQuestionResponseSerializer):
    @classmethod
    def save_question_response(
            cls, question_response_dict_list, new_parent, parent_attr_name, old_parent_responses=None):
        """

        :param question_response_dict_list:
        :param new_parent:
        :param parent_attr_name:
        :param old_parent_responses
        :return: sz_dict, err_list
                where sz_dict is (key=sz, value=(___question_response_obj, is_it_add_new, or_update_existing))
                Note: Both add_new and update_existing can be False,
                       - which means this is just a clone of archived parent obj
                    The key sz, has sz.instance object
                     - which if exists, refers to question_response associated with parent
        """
        def fill_parent_fn(data_dict):
            if data_dict:
                data_dict[parent_attr_name] = new_parent
            return new_parent

        return super().save_parent_question_response_return_changes(
            question_response_dict_list, fill_parent_fn, old_parent_responses, parent_attr_name)
